@@ -28,12 +28,11 @@
 
 #include <compiz-plugin.h>
 
-#define CORE_ABIVERSION 20080424
+#define CORE_ABIVERSION 20090619
 
 #include <stdio.h>
 #include <sys/time.h>
 
-#include <X11/Xlib-xcb.h>
 #include <X11/Xutil.h>
 #include <X11/extensions/Xdamage.h>
 #include <X11/extensions/Xcomposite.h>
@@ -72,6 +71,7 @@ COMPIZ_BEGIN_DECLS
 typedef struct _CompTexture	  CompTexture;
 typedef struct _CompIcon	  CompIcon;
 typedef struct _CompWindowExtents CompWindowExtents;
+typedef struct _CompWindowExtents CompFullscreenMonitorSet;
 typedef struct _CompProgram	  CompProgram;
 typedef struct _CompFunction	  CompFunction;
 typedef struct _CompFunctionData  CompFunctionData;
@@ -166,12 +166,6 @@ typedef struct _CompWalker        CompWalker;
 #define MwmFuncMaximize (1L << 4)
 #define MwmFuncClose    (1L << 5)
 
-#define MwmDecorHandle   (1L << 2)
-#define MwmDecorTitle    (1L << 3)
-#define MwmDecorMenu     (1L << 4)
-#define MwmDecorMinimize (1L << 5)
-#define MwmDecorMaximize (1L << 6)
-
 #define MwmDecorAll      (1L << 0)
 #define MwmDecorBorder   (1L << 1)
 #define MwmDecorHandle   (1L << 2)
@@ -191,6 +185,12 @@ typedef struct _CompWalker        CompWalker;
 #define WmMoveResizeMove             8
 #define WmMoveResizeSizeKeyboard     9
 #define WmMoveResizeMoveKeyboard    10
+#define WmMoveResizeCancel	    11
+
+/* EWMH source indication client types */
+#define ClientTypeUnknown      0
+#define ClientTypeApplication  1
+#define ClientTypePager        2
 
 #define OPAQUE 0xffff
 #define COLOR  0xffff
@@ -220,6 +220,9 @@ extern Bool       noDetection;
 extern Bool	  useDesktopHints;
 extern Bool       onlyCurrentScreen;
 
+extern char	**initialPlugins;
+extern int 	nInitialPlugins;
+
 extern int  defaultRefreshRate;
 extern char *defaultTextureFilter;
 
@@ -636,8 +639,10 @@ typedef void (*FileWatchRemovedProc) (CompCore      *core,
 
 typedef struct _CompTimeout {
     struct _CompTimeout *next;
-    int			time;
-    int			left;
+    int			minTime;
+    int			maxTime;
+    int			minLeft;
+    int			maxLeft;
     CallBackProc	callBack;
     void		*closure;
     CompTimeoutHandle   handle;
@@ -651,6 +656,10 @@ typedef struct _CompWatchFd {
     CompWatchFdHandle   handle;
 } CompWatchFd;
 
+typedef void (*LogMessageProc) (const char   *componentName,
+				CompLogLevel level,
+				const char   *message);
+
 struct _CompCore {
     CompObject base;
 
@@ -683,6 +692,7 @@ struct _CompCore {
     FileWatchRemovedProc fileWatchRemoved;
 
     SessionEventProc sessionEvent;
+    LogMessageProc   logMessage;
 };
 
 int
@@ -739,67 +749,33 @@ removeFileWatch (CompFileWatchHandle handle);
 #define COMP_DISPLAY_OPTION_AUTORAISE_DELAY                  5
 #define COMP_DISPLAY_OPTION_CLOSE_WINDOW_KEY                 6
 #define COMP_DISPLAY_OPTION_CLOSE_WINDOW_BUTTON              7
-#define COMP_DISPLAY_OPTION_MAIN_MENU_KEY                    8
-#define COMP_DISPLAY_OPTION_RUN_DIALOG_KEY                   9
-#define COMP_DISPLAY_OPTION_COMMAND0                         10
-#define COMP_DISPLAY_OPTION_COMMAND1                         11
-#define COMP_DISPLAY_OPTION_COMMAND2                         12
-#define COMP_DISPLAY_OPTION_COMMAND3                         13
-#define COMP_DISPLAY_OPTION_COMMAND4                         14
-#define COMP_DISPLAY_OPTION_COMMAND5                         15
-#define COMP_DISPLAY_OPTION_COMMAND6                         16
-#define COMP_DISPLAY_OPTION_COMMAND7                         17
-#define COMP_DISPLAY_OPTION_COMMAND8                         18
-#define COMP_DISPLAY_OPTION_COMMAND9                         19
-#define COMP_DISPLAY_OPTION_COMMAND10                        20
-#define COMP_DISPLAY_OPTION_COMMAND11                        21
-#define COMP_DISPLAY_OPTION_RUN_COMMAND0_KEY                 22
-#define COMP_DISPLAY_OPTION_RUN_COMMAND1_KEY                 23
-#define COMP_DISPLAY_OPTION_RUN_COMMAND2_KEY                 24
-#define COMP_DISPLAY_OPTION_RUN_COMMAND3_KEY                 25
-#define COMP_DISPLAY_OPTION_RUN_COMMAND4_KEY                 26
-#define COMP_DISPLAY_OPTION_RUN_COMMAND5_KEY                 27
-#define COMP_DISPLAY_OPTION_RUN_COMMAND6_KEY                 28
-#define COMP_DISPLAY_OPTION_RUN_COMMAND7_KEY                 29
-#define COMP_DISPLAY_OPTION_RUN_COMMAND8_KEY                 30
-#define COMP_DISPLAY_OPTION_RUN_COMMAND9_KEY                 31
-#define COMP_DISPLAY_OPTION_RUN_COMMAND10_KEY                32
-#define COMP_DISPLAY_OPTION_RUN_COMMAND11_KEY                33
-#define COMP_DISPLAY_OPTION_SLOW_ANIMATIONS_KEY              34
-#define COMP_DISPLAY_OPTION_RAISE_WINDOW_KEY                 35
-#define COMP_DISPLAY_OPTION_RAISE_WINDOW_BUTTON              36
-#define COMP_DISPLAY_OPTION_LOWER_WINDOW_KEY                 37
-#define COMP_DISPLAY_OPTION_LOWER_WINDOW_BUTTON              38
-#define COMP_DISPLAY_OPTION_UNMAXIMIZE_WINDOW_KEY            39
-#define COMP_DISPLAY_OPTION_MINIMIZE_WINDOW_KEY              40
-#define COMP_DISPLAY_OPTION_MINIMIZE_WINDOW_BUTTON           41
-#define COMP_DISPLAY_OPTION_MAXIMIZE_WINDOW_KEY              42
-#define COMP_DISPLAY_OPTION_MAXIMIZE_WINDOW_HORZ_KEY         43
-#define COMP_DISPLAY_OPTION_MAXIMIZE_WINDOW_VERT_KEY         44
-#define COMP_DISPLAY_OPTION_OPACITY_INCREASE_BUTTON          45
-#define COMP_DISPLAY_OPTION_OPACITY_DECREASE_BUTTON          46
-#define COMP_DISPLAY_OPTION_SCREENSHOT                       47
-#define COMP_DISPLAY_OPTION_RUN_SCREENSHOT_KEY               48
-#define COMP_DISPLAY_OPTION_WINDOW_SCREENSHOT                49
-#define COMP_DISPLAY_OPTION_RUN_WINDOW_SCREENSHOT_KEY        50
-#define COMP_DISPLAY_OPTION_WINDOW_MENU_BUTTON               51
-#define COMP_DISPLAY_OPTION_WINDOW_MENU_KEY                  52
-#define COMP_DISPLAY_OPTION_SHOW_DESKTOP_KEY                 53
-#define COMP_DISPLAY_OPTION_SHOW_DESKTOP_EDGE                54
-#define COMP_DISPLAY_OPTION_RAISE_ON_CLICK                   55
-#define COMP_DISPLAY_OPTION_AUDIBLE_BELL                     56
-#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_KEY      57
-#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_BUTTON   58
-#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_HORZ_KEY 59
-#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_VERT_KEY 60
-#define COMP_DISPLAY_OPTION_HIDE_SKIP_TASKBAR_WINDOWS        61
-#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_SHADED_KEY         62
-#define COMP_DISPLAY_OPTION_IGNORE_HINTS_WHEN_MAXIMIZED      63
-#define COMP_DISPLAY_OPTION_TERMINAL			     64
-#define COMP_DISPLAY_OPTION_RUN_TERMINAL_KEY		     65
-#define COMP_DISPLAY_OPTION_PING_DELAY			     66
-#define COMP_DISPLAY_OPTION_EDGE_DELAY                       67
-#define COMP_DISPLAY_OPTION_NUM				     68
+#define COMP_DISPLAY_OPTION_SLOW_ANIMATIONS_KEY              8
+#define COMP_DISPLAY_OPTION_RAISE_WINDOW_KEY                 9
+#define COMP_DISPLAY_OPTION_RAISE_WINDOW_BUTTON              10
+#define COMP_DISPLAY_OPTION_LOWER_WINDOW_KEY                 11
+#define COMP_DISPLAY_OPTION_LOWER_WINDOW_BUTTON              12
+#define COMP_DISPLAY_OPTION_UNMAXIMIZE_WINDOW_KEY            13
+#define COMP_DISPLAY_OPTION_MINIMIZE_WINDOW_KEY              14
+#define COMP_DISPLAY_OPTION_MINIMIZE_WINDOW_BUTTON           15
+#define COMP_DISPLAY_OPTION_MAXIMIZE_WINDOW_KEY              16
+#define COMP_DISPLAY_OPTION_MAXIMIZE_WINDOW_HORZ_KEY         17
+#define COMP_DISPLAY_OPTION_MAXIMIZE_WINDOW_VERT_KEY         18
+#define COMP_DISPLAY_OPTION_WINDOW_MENU_BUTTON               19
+#define COMP_DISPLAY_OPTION_WINDOW_MENU_KEY                  20
+#define COMP_DISPLAY_OPTION_SHOW_DESKTOP_KEY                 21
+#define COMP_DISPLAY_OPTION_SHOW_DESKTOP_EDGE                22
+#define COMP_DISPLAY_OPTION_RAISE_ON_CLICK                   23
+#define COMP_DISPLAY_OPTION_AUDIBLE_BELL                     24
+#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_KEY      25
+#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_BUTTON   26
+#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_HORZ_KEY 27
+#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_VERT_KEY 28
+#define COMP_DISPLAY_OPTION_HIDE_SKIP_TASKBAR_WINDOWS        29
+#define COMP_DISPLAY_OPTION_TOGGLE_WINDOW_SHADED_KEY         30
+#define COMP_DISPLAY_OPTION_IGNORE_HINTS_WHEN_MAXIMIZED      31
+#define COMP_DISPLAY_OPTION_PING_DELAY			     32
+#define COMP_DISPLAY_OPTION_EDGE_DELAY                       33
+#define COMP_DISPLAY_OPTION_NUM				     34
 
 typedef void (*HandleEventProc) (CompDisplay *display,
 				 XEvent	     *event);
@@ -886,18 +862,11 @@ typedef void (*MatchExpHandlerChangedProc) (CompDisplay *display);
 typedef void (*MatchPropertyChangedProc) (CompDisplay *display,
 					  CompWindow  *window);
 
-typedef void (*LogMessageProc) (CompDisplay  *d,
-				const char   *componentName,
-				CompLogLevel level,
-				const char   *message);
-
 struct _CompDisplay {
     CompObject base;
 
     CompDisplay *next;
 
-    xcb_connection_t *connection;
-
     Display    *display;
     CompScreen *screens;
 
@@ -1000,6 +969,7 @@ struct _CompDisplay {
     Atom wmUserTimeAtom;
 
     Atom wmIconAtom;
+    Atom wmIconGeometryAtom;
 
     Atom clientListAtom;
     Atom clientListStackingAtom;
@@ -1015,10 +985,12 @@ struct _CompDisplay {
     Atom wmDeleteWindowAtom;
     Atom wmTakeFocusAtom;
     Atom wmPingAtom;
-    Atom wmSyncRequestAtom;
 
+    Atom wmSyncRequestAtom;
     Atom wmSyncRequestCounterAtom;
 
+    Atom wmFullscreenMonitorsAtom;
+
     Atom closeWindowAtom;
     Atom wmMoveResizeAtom;
     Atom moveResizeWindowAtom;
@@ -1029,8 +1001,6 @@ struct _CompDisplay {
     Atom xBackgroundAtom[2];
 
     Atom toolkitActionAtom;
-    Atom toolkitActionMainMenuAtom;
-    Atom toolkitActionRunDialogAtom;
     Atom toolkitActionWindowMenuAtom;
     Atom toolkitActionForceQuitDialogAtom;
 
@@ -1090,6 +1060,8 @@ struct _CompDisplay {
     MatchPropertyChangedProc   matchPropertyChanged;
 
     LogMessageProc logMessage;
+
+    void *reserved;
 };
 
 #define GET_CORE_DISPLAY(object) ((CompDisplay *) (object))
@@ -1136,15 +1108,13 @@ setDisplayOption (CompPlugin	  *plugin,
 		  CompOptionValue *value);
 
 void
-compLogMessage (CompDisplay  *d,
-		const char   *componentName,
+compLogMessage (const char   *componentName,
 		CompLogLevel level,
 		const char   *format,
 		...);
 
 void
-logMessage (CompDisplay  *d,
-	    const char   *componentName,
+logMessage (const char   *componentName,
 	    CompLogLevel level,
 	    const char   *message);
 
@@ -1398,10 +1368,10 @@ typedef void (*ApplyScreenTransformProc) (CompScreen		  *screen,
 					  CompOutput		  *output,
 					  CompTransform	          *transform);
 
-typedef void (*EnableOutputClippingProc) (CompScreen 	      *screen,
+typedef void (*EnableOutputClippingProc) (CompScreen	      *screen,
 					  const CompTransform *transform,
 					  Region	      region,
-					  CompOutput 	      *output);
+					  CompOutput	      *output);
 
 typedef void (*DisableOutputClippingProc) (CompScreen *screen);
 
@@ -1540,7 +1510,7 @@ void
 enableOutputClipping (CompScreen	  *screen,
 		      const CompTransform *transform,
 		      Region		  region,
-		      CompOutput 	  *output);
+		      CompOutput	  *output);
 
 void
 disableOutputClipping (CompScreen *screen);
@@ -1704,21 +1674,18 @@ disableTexture (CompScreen  *screen,
 #define COMP_SCREEN_OPTION_REFRESH_RATE		  2
 #define COMP_SCREEN_OPTION_HSIZE		  3
 #define COMP_SCREEN_OPTION_VSIZE		  4
-#define COMP_SCREEN_OPTION_OPACITY_STEP		  5
-#define COMP_SCREEN_OPTION_UNREDIRECT_FS	  6
-#define COMP_SCREEN_OPTION_DEFAULT_ICON		  7
-#define COMP_SCREEN_OPTION_SYNC_TO_VBLANK	  8
-#define COMP_SCREEN_OPTION_NUMBER_OF_DESKTOPS	  9
-#define COMP_SCREEN_OPTION_DETECT_OUTPUTS	  10
-#define COMP_SCREEN_OPTION_OUTPUTS		  11
-#define COMP_SCREEN_OPTION_OVERLAPPING_OUTPUTS	  12
-#define COMP_SCREEN_OPTION_FOCUS_PREVENTION_LEVEL 13
-#define COMP_SCREEN_OPTION_FOCUS_PREVENTION_MATCH 14
-#define COMP_SCREEN_OPTION_OPACITY_MATCHES	  15
-#define COMP_SCREEN_OPTION_OPACITY_VALUES	  16
-#define COMP_SCREEN_OPTION_TEXTURE_COMPRESSION	  17
-#define COMP_SCREEN_OPTION_FORCE_INDEPENDENT      18
-#define COMP_SCREEN_OPTION_NUM		          19
+#define COMP_SCREEN_OPTION_UNREDIRECT_FS	  5
+#define COMP_SCREEN_OPTION_DEFAULT_ICON		  6
+#define COMP_SCREEN_OPTION_SYNC_TO_VBLANK	  7
+#define COMP_SCREEN_OPTION_NUMBER_OF_DESKTOPS	  8
+#define COMP_SCREEN_OPTION_DETECT_OUTPUTS	  9
+#define COMP_SCREEN_OPTION_OUTPUTS		  10
+#define COMP_SCREEN_OPTION_OVERLAPPING_OUTPUTS	  11
+#define COMP_SCREEN_OPTION_FOCUS_PREVENTION_LEVEL 12
+#define COMP_SCREEN_OPTION_FOCUS_PREVENTION_MATCH 13
+#define COMP_SCREEN_OPTION_TEXTURE_COMPRESSION	  14
+#define COMP_SCREEN_OPTION_FORCE_INDEPENDENT      15
+#define COMP_SCREEN_OPTION_NUM		          16
 
 #ifndef GLX_EXT_texture_from_pixmap
 #define GLX_BIND_TO_TEXTURE_RGB_EXT        0x20D0
@@ -1748,8 +1715,9 @@ disableTexture (CompScreen  *screen,
 
 #define FOCUS_PREVENTION_LEVEL_NONE     0
 #define FOCUS_PREVENTION_LEVEL_LOW      1
-#define FOCUS_PREVENTION_LEVEL_HIGH     2
-#define FOCUS_PREVENTION_LEVEL_VERYHIGH 3
+#define FOCUS_PREVENTION_LEVEL_NORMAL   2
+#define FOCUS_PREVENTION_LEVEL_HIGH     3
+#define FOCUS_PREVENTION_LEVEL_VERYHIGH 4
 #define FOCUS_PREVENTION_LEVEL_LAST     FOCUS_PREVENTION_LEVEL_VERYHIGH
 
 typedef void (*FuncPtr) (void);
@@ -1794,6 +1762,8 @@ typedef GLXPixmap (*GLXCreatePixmapProc) (Display     *display,
 					  GLXFBConfig config,
 					  Pixmap      pixmap,
 					  const int   *attribList);
+typedef void (*GLXDestroyPixmapProc) (Display   *display,
+				      GLXPixmap pixmap);
 
 typedef void (*GLActiveTextureProc) (GLenum texture);
 typedef void (*GLClientActiveTextureProc) (GLenum texture);
@@ -1871,7 +1841,8 @@ typedef Bool (*PlaceWindowProc) (CompWindow *window,
 
 typedef void (*ValidateWindowResizeRequestProc) (CompWindow     *window,
 						 unsigned int   *mask,
-						 XWindowChanges *xwc);
+						 XWindowChanges *xwc,
+						 unsigned int   source);
 
 typedef void (*WindowResizeNotifyProc) (CompWindow *window,
 					int        dx,
@@ -1884,10 +1855,11 @@ typedef void (*WindowMoveNotifyProc) (CompWindow *window,
 				      int	 dy,
 				      Bool	 immediate);
 
-#define CompWindowGrabKeyMask    (1 << 0)
-#define CompWindowGrabButtonMask (1 << 1)
-#define CompWindowGrabMoveMask   (1 << 2)
-#define CompWindowGrabResizeMask (1 << 3)
+#define CompWindowGrabKeyMask         (1 << 0)
+#define CompWindowGrabButtonMask      (1 << 1)
+#define CompWindowGrabMoveMask        (1 << 2)
+#define CompWindowGrabResizeMask      (1 << 3)
+#define CompWindowGrabExternalAppMask (1 << 4)
 
 typedef void (*WindowGrabNotifyProc) (CompWindow   *window,
 				      int	   x,
@@ -1902,6 +1874,10 @@ typedef void (*WindowStateChangeNotifyProc) (CompWindow   *window,
 
 typedef void (*OutputChangeNotifyProc) (CompScreen *screen);
 
+typedef unsigned int (*AddSupportedAtomsProc) (CompScreen   *s,
+					       Atom         *atoms,
+					       unsigned int size);
+
 typedef void (*InitWindowWalkerProc) (CompScreen *screen,
 				      CompWalker *walker);
 
@@ -2163,6 +2139,7 @@ struct _CompScreen {
     GLXGetFBConfigsProc      getFBConfigs;
     GLXGetFBConfigAttribProc getFBConfigAttrib;
     GLXCreatePixmapProc      createPixmap;
+    GLXDestroyPixmapProc     destroyPixmap;
 
     GLActiveTextureProc       activeTexture;
     GLClientActiveTextureProc clientActiveTexture;
@@ -2221,8 +2198,11 @@ struct _CompScreen {
     WindowStateChangeNotifyProc windowStateChangeNotify;
 
     OutputChangeNotifyProc outputChangeNotify;
+    AddSupportedAtomsProc  addSupportedAtoms;
 
     InitWindowWalkerProc initWindowWalker;
+
+    void *reserved;
 };
 
 #define GET_CORE_SCREEN(object) ((CompScreen *) (object))
@@ -2277,6 +2257,9 @@ void
 setCurrentOutput (CompScreen *s,
 		  int	     outputNum);
 
+void
+setSupportedWmHints (CompScreen *s);
+
 void
 updateScreenBackground (CompScreen  *screen,
 			CompTexture *texture);
@@ -2543,6 +2526,12 @@ typedef enum {
     CompStackingUpdateModeInitialMapDeniedFocus
 } CompStackingUpdateMode;
 
+typedef enum {
+    CompFocusAllowed = 0,
+    CompFocusPrevent,
+    CompFocusDenied
+} CompFocusResult;
+
 struct _CompWindowExtents {
     int left;
     int right;
@@ -2577,6 +2566,7 @@ struct _CompWindow {
     int		      serverBorderWidth;
     Window	      transientFor;
     Window	      clientLeader;
+    XWMHints	      *hints;
     XSizeHints	      sizeHints;
     Pixmap	      pixmap;
     CompTexture       *texture;
@@ -2632,13 +2622,6 @@ struct _CompWindow {
     unsigned int lastPong;
     Bool	 alive;
 
-    GLushort opacity;
-    GLushort brightness;
-    GLushort saturation;
-
-    Bool opacityPropSet;
-    int  opacityFactor;
-
     WindowPaintAttrib paint;
     WindowPaintAttrib lastPaint;
 
@@ -2652,6 +2635,12 @@ struct _CompWindow {
     CompIcon **icon;
     int	     nIcon;
 
+    XRectangle iconGeometry;
+    Bool       iconGeometrySet;
+
+    XRectangle fullscreenMonitorRect;
+    Bool       fullscreenMonitorsSet;
+
     XWindowChanges saveWc;
     int		   saveMask;
 
@@ -2687,6 +2676,8 @@ struct _CompWindow {
 
     /* must be set by addWindowGeometry */
     DrawWindowGeometryProc drawWindowGeometry;
+
+    void *reserved;
 };
 
 #define GET_CORE_WINDOW(object) ((CompWindow *) (object))
@@ -2799,9 +2790,6 @@ setWindowProp32 (CompDisplay    *display,
 		 Atom		property,
 		 unsigned short value);
 
-void
-updateWindowOpacity (CompWindow *window);
-
 void
 updateNormalHints (CompWindow *window);
 
@@ -2814,6 +2802,9 @@ updateWindowClassHints (CompWindow *window);
 void
 updateTransientHint (CompWindow *w);
 
+void
+updateIconGeometry (CompWindow *w);
+
 Window
 getClientLeader (CompWindow *w);
 
@@ -2829,6 +2820,10 @@ setWmState (CompDisplay *display,
 	    int		state,
 	    Window      id);
 
+void
+setWindowFullscreenMonitors (CompWindow               *w,
+			     CompFullscreenMonitorSet *monitors);
+
 void
 setWindowFrameExtents (CompWindow	 *w,
 		       CompWindowExtents *input);
@@ -2884,13 +2879,15 @@ unsigned int
 adjustConfigureRequestForGravity (CompWindow     *w,
 				  XWindowChanges *xwc,
 				  unsigned int   xwcm,
-				  int            gravity);
+				  int            gravity,
+				  int            direction);
 
 void
 moveResizeWindow (CompWindow     *w,
 		  XWindowChanges *xwc,
 		  unsigned int   xwcm,
-		  int            gravity);
+		  int            gravity,
+		  unsigned int   source);
 
 void
 syncWindowPosition (CompWindow *w);
@@ -2959,7 +2956,8 @@ placeWindow (CompWindow *w,
 void
 validateWindowResizeRequest (CompWindow     *w,
 			     unsigned int   *mask,
-			     XWindowChanges *xwc);
+			     XWindowChanges *xwc,
+			     unsigned int   source);
 
 void
 windowResizeNotify (CompWindow *w,
@@ -3050,9 +3048,11 @@ void
 setWindowUserTime (CompWindow *w,
 		   Time       time);
 
-Bool
+CompFocusResult
 allowWindowFocus (CompWindow   *w,
 		  unsigned int noFocusMask,
+		  unsigned int viewportX,
+		  unsigned int viewportY,
 		  Time         timestamp);
 
 void
