@@ -61,16 +61,9 @@
 #define DBUS_METHOD_GET     "get"
 #define DBUS_SIGNAL_CHANGED "changed"
 
-double decorationOpacity = 0.75;
-bool   decorationOpacityShade = false;
-double activeDecorationOpacity = 1.0;
-bool   activeDecorationOpacityShade = false;
 int    blurType = BLUR_TYPE_NONE;
 
-decor_context_t KWD::Decorator::mDefaultContext;
-decor_extents_t KWD::Decorator::mDefaultBorder;
 decor_shadow_t *KWD::Decorator::mNoBorderShadow = 0;
-decor_shadow_t *KWD::Decorator::mDefaultShadow  = 0;
 KWD::PluginManager *KWD::Decorator::mPlugins = 0;
 KWD::Options *KWD::Decorator::mOptions = 0;
 NETRootInfo *KWD::Decorator::mRootInfo;
@@ -91,10 +84,8 @@ KWD::PluginManager::PluginManager (KSharedConfigPtr config):
 }
 
 
-KWD::Decorator::Decorator (Display* display,
-			   Qt::HANDLE visual,
-			   Qt::HANDLE colormap) :
-    KApplication (display, visual, colormap),
+KWD::Decorator::Decorator () :
+    KApplication (),
     mConfig (0),
     mCompositeWindow (0),
     mSwitcher (0)
@@ -200,13 +191,11 @@ KWD::Decorator::~Decorator (void)
 }
 
 bool
-KWD::Decorator::enableDecorations (Time timestamp,
-				   int  damageEvent)
+KWD::Decorator::enableDecorations (Time timestamp)
 {
     QList <WId>::ConstIterator it;
 
     mDmSnTimestamp = timestamp;
-    mDamageEvent   = damageEvent;
 
     if (!pluginManager ()->loadPlugin (""))
 	return false;
@@ -239,8 +228,6 @@ KWD::Decorator::enableDecorations (Time timestamp,
     foreach (WId id, KWindowSystem::windows ())
 	handleWindowAdded (id);
 
-    connect (&mIdleTimer, SIGNAL (timeout ()), SLOT (processDamage ()));
-
     connect (Plasma::Theme::defaultTheme (), SIGNAL (themeChanged ()),
 	     SLOT (plasmaThemeChanged ()));
 
@@ -251,31 +238,6 @@ KWD::Decorator::enableDecorations (Time timestamp,
     return true;
 }
 
-void
-KWD::Decorator::updateDefaultShadow (KWD::Window *w)
-{
-    bool uniqueHorzShape, uniqueVertShape;
-
-    if (mDefaultShadow)
-    {
-	decor_shadow_destroy (QX11Info::display(), mDefaultShadow);
-	mDefaultShadow = NULL;
-    }
-
-    w->getShapeInfo (&uniqueHorzShape, &uniqueVertShape);
-
-    /* only return shadow if decoration doesn't use a unique shape */
-    if (uniqueHorzShape || uniqueVertShape)
-	return;
-
-    mDefaultContext = *w->context ();
-    mDefaultBorder  = *w->border ();
-    mDefaultShadow  = w->shadow ();
-
-    if (mDefaultShadow)
-	decor_shadow_reference (mDefaultShadow);
-}
-
 void
 KWD::Decorator::updateAllShadowOptions (void)
 {
@@ -345,12 +307,6 @@ KWD::Decorator::changeShadowOptions (decor_shadow_options_t *opt)
     mShadowOptions = *opt;
 
     updateShadow ();
-
-    mDecorNormal->reloadDecoration ();
-    mDecorActive->reloadDecoration ();
-
-    for (it = mClients.constBegin (); it != mClients.constEnd (); it++)
-	it.value ()->reloadDecoration ();
 }
 
 void
@@ -362,12 +318,6 @@ KWD::Decorator::updateShadow (void)
 
     xscreen = ScreenOfDisplay (xdisplay, QX11Info::appScreen ());
 
-    if (mDefaultShadow)
-    {
-	decor_shadow_destroy (xdisplay, mDefaultShadow);
-	mDefaultShadow = NULL;
-    }
-
     if (mNoBorderShadow)
 	decor_shadow_destroy (xdisplay, mNoBorderShadow);
 
@@ -410,69 +360,19 @@ KWD::Decorator::updateShadow (void)
     }
 }
 
-void
-KWD::Decorator::processDamage (void)
-{
-    QMap <WId, KWD::Window *>::ConstIterator it;
-
-    mDecorNormal->processDamage ();
-    mDecorActive->processDamage ();
-
-    for (it = mClients.constBegin (); it != mClients.constEnd (); it++)
-	it.value ()->processDamage ();
-}
-
 bool
 KWD::Decorator::x11EventFilter (XEvent *xevent)
 {
     KWD::Window *client;
     int		status;
-
+    
     switch (xevent->type) {
-    case MapNotify: {
-	XMapEvent *xme = reinterpret_cast <XMapEvent *> (xevent);
-
-	if (mWindows.contains (xme->window))
-	    client = mWindows[xme->window];
-	else if (mDecorNormal->winId () == xme->window)
-	    client = mDecorNormal;
-	else if (mDecorActive->winId () == xme->window)
-	    client = mDecorActive;
-	else
-	    break;
-
-	if (client->handleMap ())
-	{
-	    if (!mIdleTimer.isActive ())
-	    {
-		mIdleTimer.setSingleShot (true);
-		mIdleTimer.start (0);
-	    }
-	}
-    } break;
     case ConfigureNotify: {
 	XConfigureEvent *xce = reinterpret_cast <XConfigureEvent *> (xevent);
 
 	if (mFrames.contains (xce->window))
 	    mFrames[xce->window]->updateFrame (xce->window);
 
-	if (mWindows.contains (xce->window))
-	    client = mWindows[xce->window];
-	else if (mDecorNormal->winId () == xce->window)
-	    client = mDecorNormal;
-	else if (mDecorActive->winId () == xce->window)
-	    client = mDecorActive;
-	else
-	    break;
-
-	if (client->handleConfigure (QSize (xce->width, xce->height)))
-	{
-	    if (!mIdleTimer.isActive ())
-	    {
-		mIdleTimer.setSingleShot (true);
-		mIdleTimer.start (0);
-	    }
-	}
     } break;
     case SelectionRequest:
 	decor_handle_selection_request (QX11Info::display(), xevent, mDmSnTimestamp);
@@ -513,6 +413,9 @@ KWD::Decorator::x11EventFilter (XEvent *xevent)
 
 	client = mFrames[xce->window];
 
+	if (!client->decorWidget ())
+	    break;
+
 	child = client->childAt (xce->x, xce->y);
 	if (child)
 	{
@@ -550,6 +453,9 @@ KWD::Decorator::x11EventFilter (XEvent *xevent)
 
 	client = mFrames[xme->window];
 
+	if (!client->decorWidget ())
+	    break;
+
 	child = client->childAt (xme->x, xme->y);
 
 	if (child)
@@ -569,8 +475,8 @@ KWD::Decorator::x11EventFilter (XEvent *xevent)
 		client->setActiveChild (child);
 	    }
 
-	    if (client != child)
-		qp -= QPoint (child->pos ().x (), child->pos ().y ());
+	    if (client->decorWidget () != child)
+		qp = child->mapFrom (client->decorWidget (), qp);
 
 	    QMouseEvent qme (QEvent::MouseMove, qp, Qt::NoButton,
 			     Qt::NoButton, Qt::NoModifier);
@@ -591,17 +497,25 @@ KWD::Decorator::x11EventFilter (XEvent *xevent)
 
 	client = mFrames[xbe->window];
 
+	if (!client->decorWidget ())
+	    break;
+
 	child = client->childAt (xbe->x, xbe->y);
 
 	if (child)
 	{
 	    XButtonEvent xbe2 = *xbe;
 	    xbe2.window = child->winId ();
-	    if (client != child)
-	    {
-		xbe2.x = xbe->x - child->pos ().x ();
-		xbe2.y = xbe->y - child->pos ().y ();
-	    }
+
+	    QPoint p;
+		
+	    p = client->mapToChildAt (QPoint (xbe->x, xbe->y));
+	    xbe2.x = p.x ();
+	    xbe2.y = p.y ();
+	    
+	    p = child->mapToGlobal(p);
+	    xbe2.x_root = p.x ();
+	    xbe2.y_root = p.y ();
 
 	    client->setFakeRelease (false);
 	    QApplication::x11ProcessEvent ((XEvent *) &xbe2);
@@ -661,36 +575,6 @@ KWD::Decorator::x11EventFilter (XEvent *xevent)
 	}
 	break;
     default:
-	if (xevent->type == mDamageEvent + XDamageNotify)
-	{
-	    XDamageNotifyEvent *xde =
-		reinterpret_cast <XDamageNotifyEvent *>(xevent);
-
-	    if (mWindows.contains (xde->drawable))
-		client = mWindows[xde->drawable];
-	    else if (mDecorNormal->winId () == xde->drawable)
-		client = mDecorNormal;
-	    else if (mDecorActive->winId () == xde->drawable)
-		client = mDecorActive;
-	    else
-		break;
-
-	    client->addDamageRect (xde->area.x,
-				   xde->area.y,
-				   xde->area.width,
-				   xde->area.height);
-
-	    if (client->pixmapId ())
-	    {
-		if (!mIdleTimer.isActive ())
-		{
-		    mIdleTimer.setSingleShot (true);
-		    mIdleTimer.start (0);
-		}
-	    }
-
-	    return true;
-	}
 	break;
     }
 
@@ -794,7 +678,6 @@ KWD::Decorator::handleWindowAdded (WId id)
 				      height + border * 2);
 
 	    mClients.insert (id, client);
-	    mWindows.insert (client->winId (), client);
 	    mFrames.insert (frame, client);
 	}
 	else
@@ -814,7 +697,6 @@ KWD::Decorator::handleWindowAdded (WId id)
 	if (client)
 	{
 	    mClients.remove (client->windowId ());
-	    mWindows.remove (client->winId ());
 	    mFrames.remove (client->frameId ());
 
 	    delete client;
@@ -835,7 +717,6 @@ KWD::Decorator::handleWindowRemoved (WId id)
     if (window)
     {
 	mClients.remove (window->windowId ());
-	mWindows.remove (window->winId ());
 	mFrames.remove (window->frameId ());
 	delete window;
     }
