@@ -1,5 +1,5 @@
 /*
- * Copyright © 2008 Dennis Kasprzyk <onestone@opencompositing.org>
+ * Copyright © 2009 Dennis Kasprzyk <onestone@compiz-fusion.org>
  * Copyright © 2006 Novell, Inc.
  * Copyright © 2006 Volker Krause <vkrause@kde.org>
  *
@@ -55,6 +55,9 @@
 #include <QVector>
 #include <QProcess>
 #include <QStyle>
+#include <QPainter>
+
+#include "paintredirector.h"
 
 KWD::Window::Window (WId  parentId,
 		     WId  clientId,
@@ -63,7 +66,7 @@ KWD::Window::Window (WId  parentId,
 		     int  x,
 		     int  y,
 		     int  w,
-		     int  h): QWidget (0, 0),
+		     int  h) :
     mType (type),
     mParentId (parentId),
     mFrame (0),
@@ -71,22 +74,12 @@ KWD::Window::Window (WId  parentId,
     mSelectedId (0),
     mDecor (0),
     mPixmap (0),
-    mDamageId (0),
-    mShadow (0),
-    mPicture (0),
-    mTexturePicture (0),
-    mDecorationPicture (0),
     mUpdateProperty (false),
     mShapeSet (false),
-    mUniqueHorzShape (false),
-    mUniqueVertShape (false),
     mPopup (0),
     mAdvancedMenu (0),
     mOpacityMenu (0),
     mDesktopMenu (0),
-    mMapped (false),
-    mPendingMap (0),
-    mPendingConfigure (0),
     mProcessKiller (this),
     mKeys (this),
     mResizeOpAction (0),
@@ -98,7 +91,8 @@ KWD::Window::Window (WId  parentId,
     mFullScreenOpAction (0),
     mMinimizeOpAction (0),
     mCloseOpAction (0),
-    mDesktopOpAction (0)
+    mDesktopOpAction (0),
+    mPaintRedirector (0)
 {
     memset (&mBorder, 0, sizeof (mBorder));
 
@@ -162,27 +156,18 @@ KWD::Window::Window (WId  parentId,
 
 KWD::Window::~Window (void)
 {
-    if (mShadow)
-	decor_shadow_destroy (QX11Info::display(), mShadow);
-
-    if (mPicture)
-	XRenderFreePicture (QX11Info::display(), mPicture);
-
     if (mPixmap)
 	XFreePixmap (QX11Info::display(), mPixmap);
 
-    if (mTexturePicture)
-	XRenderFreePicture (QX11Info::display(), mTexturePicture);
-
-    if (mDecorationPicture)
-	XRenderFreePicture (QX11Info::display(), mDecorationPicture);
-
     if (mDecor)
 	delete mDecor;
 
     if (mPopup)
 	delete mPopup;
 
+    if (mPaintRedirector)
+	delete mPaintRedirector;
+
     if (mProcessKiller.state () == QProcess::Running)
     {
 	mProcessKiller.terminate ();
@@ -372,8 +357,6 @@ KWD::Window::caption (void) const
 void
 KWD::Window::showWindowMenu (const QPoint &pos)
 {
-    QPoint pnt;
-
     if (!mPopup)
     {
 	QAction *action;
@@ -489,11 +472,11 @@ KWD::Window::showWindowMenu (const QPoint &pos)
         mCloseOpAction->setData (KDecorationDefines::CloseOp);
     }
 
-    pnt = mapFromGlobal (pos);
+    QPoint pnt = mDecor->widget ()->mapFromGlobal (pos);
+
+    pnt += QPoint (mGeometry.x () - mBorder.left - mPadding.left,
+		   mGeometry.y () - mBorder.top - mPadding.top);
 
-    pnt += QPoint (mGeometry.x () - mBorder.left,
-		   mGeometry.y () - mBorder.top);
- 
     mPopup->exec (pnt);
 }
 
@@ -606,7 +589,7 @@ KWD::Window::isPreview (void) const
 QRect
 KWD::Window::geometry (void) const
 {
-    QRect rect = QWidget::geometry ();
+    QRect rect = mGeometry;
 
     return QRect (rect.x () - ROOT_OFF_X,
 		  rect.y () - ROOT_OFF_Y,
@@ -749,7 +732,7 @@ KWD::Window::currentDesktop (void) const
 QWidget *
 KWD::Window::initialParentWidget (void) const
 {
-    return const_cast <Window *> (this);
+    return 0;
 }
 
 Qt::WFlags
@@ -763,31 +746,12 @@ KWD::Window::grabXServer (bool)
 {
 }
 
-void
-KWD::Window::repaintShadow (void)
-{
-}
-
 bool
 KWD::Window::compositingActive (void) const
 {
     return true;
 }
 
-bool
-KWD::Window::shadowsActive (void) const
-{
-    /* we are drawing the shadows ourselves, no need for the
-       decoration engine to do so */
-    return false;
-}
-
-double
-KWD::Window::opacity (void) const
-{
-    return 1.0;
-}
-
 void
 KWD::Window::createDecoration (void)
 {
@@ -800,6 +764,21 @@ KWD::Window::createDecoration (void)
     decor->init ();
 
     mDecor = decor;
+    
+    mDecor->widget ()->installEventFilter (this);
+
+    mPaintRedirector = new KWin::PaintRedirector (mDecor->widget ());
+    connect (mPaintRedirector, SIGNAL (paintPending()),
+	     this, SLOT (decorRepaintPending ()));
+
+    mPadding.top = mPadding.bottom = mPadding.left = mPadding.right = 0;
+
+    if (KDecorationUnstable *deco2 = dynamic_cast<KDecorationUnstable*>(decor))
+        deco2->padding (mPadding.left, mPadding.right, mPadding.top, mPadding.bottom);
+
+    XReparentWindow (QX11Info::display(), mDecor->widget ()->winId (), mParentId, 0, 0);
+    
+    //decor->widget()->move(-mPadding.left, -mPadding.top);
 
     if (mType == Normal && mFrame)
     {
@@ -812,356 +791,30 @@ KWD::Window::createDecoration (void)
 	    return;
     }
 
-    KWD::trapXError ();
-    XSelectInput (QX11Info::display(), this->winId(),
-		  StructureNotifyMask | PropertyChangeMask);
-    KWD::popXError ();
-
     resizeDecoration (true);
 }
 
-static void
-fillQRegion (Display *xdisplay,
-	     Picture picture,
-	     int     clipX1,
-	     int     clipY1,
-	     int     clipX2,
-	     int     clipY2,
-	     int     xOff,
-	     int     yOff,
-	     QRegion *region)
-{
-    static XRenderColor		   white = { 0xffff, 0xffff, 0xffff, 0xffff };
-    QVector <QRect>		   rects = region->rects ();
-    int				   x1, y1, x2, y2;
-
-    foreach (QRect rect, rects)
-    {
-	x1 = rect.x ();
-	y1 = rect.y ();
-	x2 = x1 + rect.width ();
-	y2 = y1 + rect.height ();
-
-	if (x1 < clipX1)
-	    x1 = clipX1;
-	if (y1 < clipY1)
-	    y1 = clipY1;
-	if (x2 > clipX2)
-	    x2 = clipX2;
-	if (y2 > clipY2)
-	    y2 = clipY2;
-
-	if (x1 < x2 && y1 < y2)
-	    XRenderFillRectangle (xdisplay, PictOpSrc, picture, &white,
-				  xOff + x1,
-				  yOff + y1,
-				  x2 - x1,
-				  y2 - y1);
-    }
-}
-
-static void
-drawBorderShape (Display	 *xdisplay,
-		 Pixmap		 pixmap,
-		 Picture	 picture,
-		 int		 width,
-		 int		 height,
-		 decor_context_t *c,
-		 void		 *closure)
-{
-    static XRenderColor clear = { 0x0000, 0x0000, 0x0000, 0x0000 };
-    static XRenderColor white = { 0xffff, 0xffff, 0xffff, 0xffff };
-    KWD::Window		*w = (KWD::Window *) closure;
-    QRegion		*shape;
-    bool		uniqueHorzShade;
-    bool		uniqueVertShade;
-    int			xOffLeft, yOffTop, xOffRight, yOffBottom;
-    QRect		rect = w->geometry ();
-    int			x1, y1, x2, y2;
-
-    (void) pixmap;
-
-    XRenderFillRectangle (xdisplay, PictOpSrc, picture, &clear,
-			  0, 0, width, height);
-
-    shape = w->getShape ();
-    w->getShapeInfo (&uniqueHorzShade, &uniqueVertShade);
-
-    xOffLeft = c->left_space - c->extents.left;
-    yOffTop  = c->top_space  - c->extents.top;
-
-    xOffRight  = c->left_space - c->extents.left;
-    yOffBottom = c->top_space  - c->extents.top;
-
-    x1 = c->left_space;
-    y1 = c->top_space;
-    x2 = width - c->right_space;
-    y2 = height - c->bottom_space;
-
-    if (shape)
-    {
-	if (uniqueHorzShade && uniqueVertShade)
-	{
-	    fillQRegion (xdisplay, picture,
-			 0, 0,
-			 rect.width (), rect.height (),
-			 xOffLeft, yOffTop, shape);
-	}
-	else
-	{
-	    if (!uniqueHorzShade)
-		xOffRight = x2 - (rect.width () - c->extents.right);
-
-	    if (!uniqueVertShade)
-		yOffBottom = y2 - (rect.height () - c->extents.bottom);
-
-	    if (uniqueHorzShade)
-	    {
-		fillQRegion (xdisplay, picture,
-			     0, 0,
-			     rect.width (), c->extents.top,
-			     xOffLeft, yOffTop, shape);
-		fillQRegion (xdisplay, picture,
-			     0, rect.height () - c->extents.bottom,
-			     rect.width (), rect.height (),
-			     xOffLeft, yOffBottom, shape);
-	    }
-	    else
-	    {
-		fillQRegion (xdisplay, picture,
-			     0, 0,
-			     c->extents.left, c->extents.top,
-			     xOffLeft, yOffTop, shape);
-		fillQRegion (xdisplay, picture,
-			     rect.width () - c->extents.right, 0,
-			     rect.width (), c->extents.top,
-			     xOffRight, yOffTop, shape);
-		fillQRegion (xdisplay, picture,
-			     0, rect.height () - c->extents.bottom,
-			     c->extents.left, rect.height (),
-			     xOffLeft, yOffBottom, shape);
-		fillQRegion (xdisplay, picture,
-			     rect.width () - c->extents.right,
-			     rect.height () - c->extents.bottom,
-			     rect.width (), rect.height (),
-			     xOffRight, yOffBottom, shape);
-
-		y1 -= c->extents.top;
-		y2 += c->extents.bottom;
-	    }
-
-	    if (uniqueVertShade)
-	    {
-		fillQRegion (xdisplay, picture,
-			     0, c->extents.top,
-			     c->extents.left,
-			     rect.height () - c->extents.bottom,
-			     xOffLeft, yOffTop, shape);
-		fillQRegion (xdisplay, picture,
-			     rect.width () - c->extents.right, c->extents.top,
-			     rect.width (),
-			     rect.height () - c->extents.bottom,
-			     xOffRight, yOffTop, shape);
-	    }
-	    else
-	    {
-		x1 -= c->extents.left;
-		x2 += c->extents.right;
-	    }
-	}
-    }
-    else
-    {
-	x1 -= c->extents.left;
-	x2 += c->extents.right;
-	y1 -= c->extents.top;
-	y2 += c->extents.bottom;
-    }
-
-    XRenderFillRectangle (xdisplay, PictOpSrc, picture, &white,
-			  x1,
-			  y1,
-			  x2 - x1,
-			  y2 - y1);
-}
-
-static void
-cornersFromQRegion (QRegion *region,
-		    int     width,
-		    int     height,
-		    int     left,
-		    int     right,
-		    int     top,
-		    int     bottom,
-		    int	    *leftCorner,
-		    int	    *rightCorner,
-		    int	    *topCorner,
-		    int	    *bottomCorner)
-{
-    QRegion l, r, t, b;
-
-    l = QRegion (0, top, left, height - top - bottom) - *region;
-    r = QRegion (width - right, top, right, height - top - bottom) - *region;
-    t = QRegion (0, 0, width, top) - *region;
-    b = QRegion (0, height - bottom, width, bottom) - *region;
-
-    if (l.isEmpty ())
-	*leftCorner = left;
-    else
-	*leftCorner = left -
-	    (l.boundingRect ().x () + l.boundingRect ().width ());
-
-    if (r.isEmpty ())
-	*rightCorner = right;
-    else
-	*rightCorner = r.boundingRect ().x () - width + right;
-
-    if (t.isEmpty ())
-	*topCorner = top;
-    else
-	*topCorner = top -
-	    (t.boundingRect ().y () + t.boundingRect ().height ());
-
-    if (b.isEmpty ())
-	*bottomCorner = bottom;
-    else
-	*bottomCorner = b.boundingRect ().y () - height + bottom;
-}
-
 void
-KWD::Window::updateShadow (void)
+KWD::Window::setMask (const QRegion &region, int)
 {
-    Display	      *xdisplay = QX11Info::display();
-    Screen	      *xscreen;
-    XRenderPictFormat *xformat;
-    int		      leftCorner, rightCorner, topCorner, bottomCorner;
-
-    xscreen = ScreenOfDisplay (xdisplay, QX11Info::appScreen ());
-
-    if (mShadow)
+    if (region.isEmpty ())
     {
-	decor_shadow_destroy (QX11Info::display(), mShadow);
-	mShadow = NULL;
+      mShapeSet = false;
+      return;
     }
 
-    if (mShapeSet)
-    {
-	cornersFromQRegion (&mShape,
-			    mGeometry.width () + mBorder.left + mBorder.right,
-			    mGeometry.height () + mBorder.top + mBorder.bottom,
-			    mBorder.left,
-			    mBorder.right,
-			    mBorder.top,
-			    mBorder.bottom,
-			    &leftCorner,
-			    &rightCorner,
-			    &topCorner,
-			    &bottomCorner);
-    }
-    else
-    {
-	leftCorner   = mBorder.left;
-	rightCorner  = mBorder.right;
-	topCorner    = mBorder.top;
-	bottomCorner = mBorder.bottom;
-    }
-
-    /* use default shadow if such exist */
-    if (!mUniqueHorzShape && !mUniqueVertShape)
-    {
-	mShadow = Decorator::defaultWindowShadow (&mContext, &mBorder);
-	if (mShadow)
-	    decor_shadow_reference (mShadow);
-    }
-
-    if (!mShadow)
-    {
-	mShadow = decor_shadow_create (xdisplay,
-				       xscreen,
-				       mUniqueHorzShape ?
-				       mGeometry.width () : 1,
-				       mUniqueVertShape ?
-				       mGeometry.height () : 1,
-				       mBorder.left,
-				       mBorder.right,
-				       mBorder.top,
-				       mBorder.bottom,
-				       leftCorner,
-				       rightCorner,
-				       topCorner,
-				       bottomCorner,
-				       KWD::Decorator::shadowOptions (),
-				       &mContext,
-				       drawBorderShape,
-				       (void *) this);
-
-	if (mType == Default)
-	    KWD::Decorator::updateDefaultShadow (this);
-    }
-
-    /* create new layout */
-    if (mType == Normal)
-	decor_get_best_layout (&mContext,
-			       mGeometry.width (),
-			       mGeometry.height (),
-			       &mLayout);
-    else
-	decor_get_default_layout (&mContext,
-				  mGeometry.width (),
-				  mGeometry.height (),
-				  &mLayout);
-
-    if (mDecorationPicture)
-	XRenderFreePicture (QX11Info::display(), mDecorationPicture);
-
-    if (mTexturePicture)
-	XRenderFreePicture (QX11Info::display(), mTexturePicture);
-
-    mTexturePixmap       = QPixmap (mLayout.width, mLayout.height);
-    mTexturePixmapBuffer = QPixmap (mLayout.width, mLayout.height);
-
-    xformat = XRenderFindStandardFormat (QX11Info::display(),
-					 PictStandardARGB32);
-
-    mDecorationPicture =
-	XRenderCreatePicture (QX11Info::display(),
-			      mTexturePixmap.handle (),
-			      xformat, 0, NULL);
-    mTexturePicture =
-	XRenderCreatePicture (QX11Info::display(),
-			      mTexturePixmapBuffer.handle (),
-			      xformat, 0, NULL);
-
-    decor_fill_picture_extents_with_shadow (QX11Info::display(),
-					    mShadow,
-					    &mContext,
-					    mTexturePicture,
-					    &mLayout);
-
-    if (mPixmap)
-	mDecor->widget ()->repaint ();
-
-    mUpdateProperty = true;
-}
-
-void
-KWD::Window::setMask (const QRegion &reg, int)
-{
-    QRegion top, bottom, left, right;
-    bool    uniqueHorzShape, uniqueVertShape;
-
-    if (mShapeSet && reg == mShape)
+    if (mShapeSet && region == mShape)
 	return;
 
-    mShape    = reg;
+    mShape    = region;
     mShapeSet = true;
 
     if (mFrame)
     {
-	QRegion r;
+	QRegion r = region.translated (-mPadding.left, -mPadding.top);
 
-	r = reg - QRegion (mBorder.left, mBorder.top,
-			   mGeometry.width (), mGeometry.height ());
+	r -= QRegion (mBorder.left, mBorder.top,
+		      mGeometry.width (), mGeometry.height ());
 
 	KWD::trapXError ();
 	XShapeCombineRegion (QX11Info::display(),
@@ -1173,94 +826,36 @@ KWD::Window::setMask (const QRegion &reg, int)
 			     ShapeSet);
 	KWD::popXError ();
     }
-
-    top    = QRegion (mBorder.left, 0,
-		      mGeometry.width (), mBorder.top) - reg;
-    bottom = QRegion (mBorder.left, mGeometry.height () + mBorder.top,
-		      mGeometry.width (), mBorder.bottom) - reg;
-    left   = QRegion (0, mBorder.top, mBorder.left,
-		      mGeometry.height ()) - reg;
-    right  = QRegion (mBorder.left + mGeometry.width (), mBorder.top,
-		      mBorder.right, mGeometry.height ()) - reg;
-
-    uniqueHorzShape = !top.isEmpty ()  || !bottom.isEmpty ();
-    uniqueVertShape = !left.isEmpty () || !right.isEmpty ();
-
-    if (uniqueHorzShape || mUniqueHorzShape ||
-	uniqueVertShape || mUniqueVertShape)
-    {
-	mUniqueHorzShape = uniqueHorzShape;
-	mUniqueVertShape = uniqueVertShape;
-
-	if (mPixmap)
-	    QTimer::singleShot (0, this, SLOT (updateShadow ()));
-    }
 }
 
-bool
+void
 KWD::Window::resizeDecoration (bool force)
 {
     int w, h;
 
     mDecor->borders (mBorder.left, mBorder.right, mBorder.top, mBorder.bottom);
-
-    w = mGeometry.width () + mBorder.left + mBorder.right;
-    h = mGeometry.height () + mBorder.top + mBorder.bottom;
-
-    if (!force)
-    {
-	if (w == width () && h == height ())
-	    return FALSE;
-    }
-
-    /* reset shape */
-    mShapeSet        = false;
-    mUniqueHorzShape = false;
-    mUniqueVertShape = false;
+    
+    mExtents.left   = mBorder.left + mPadding.left;
+    mExtents.right  = mBorder.right + mPadding.right;
+    mExtents.top    = mBorder.top + mPadding.top;
+    mExtents.bottom = mBorder.bottom + mPadding.bottom;
 
     if (mType != Normal)
     {
-	Display		*xdisplay = QX11Info::display();
-	Screen		*xscreen;
-	decor_shadow_t  *tmpShadow;
-	decor_context_t c;
-
-	xscreen = ScreenOfDisplay (xdisplay, QX11Info::appScreen ());
-
-	/* XXX: we have to create a temporary shadow to get the client
-	   geometry. libdecoration should be fixed so it's able to just
-	   fill out a context struct and not necessarily generate a
-	   shadow for this purpose. */
-	tmpShadow = decor_shadow_create (xdisplay,
-					 xscreen,
-					 1, 1,
-					 mBorder.left,
-					 mBorder.right,
-					 mBorder.top,
-					 mBorder.bottom,
-					 mBorder.left,
-					 mBorder.right,
-					 mBorder.top,
-					 mBorder.bottom,
-					 KWD::Decorator::shadowOptions (),
-					 &c,
-					 decor_draw_simple,
-					 (void *) 0);
-
-	decor_shadow_destroy (xdisplay, tmpShadow);
-
-	w = c.left_corner_space + 1 + c.right_corner_space;
-
-	/* most styles render something useful at least 30 px width */
-	if (w < 30)
-	    w = 30;
-
-	mGeometry = QRect (50, 50, w,
-			   c.top_corner_space + 1 + c.bottom_corner_space);
+	mGeometry = QRect (50, 50, 100, 100);
     }
 
-    w = mGeometry.width () + mBorder.left + mBorder.right;
-    h = mGeometry.height () + mBorder.top + mBorder.bottom;
+    w = mGeometry.width () + mExtents.left + mExtents.right;
+    h = mGeometry.height () + mExtents.top + mExtents.bottom;
+    
+    if (!force)
+    {
+      if (w == decorWidget ()->width () && h == decorWidget ()->height ())
+        return;
+    }
+    
+    /* reset shape */
+    mShapeSet        = false;
 
     if (mPixmap)
     {
@@ -1268,106 +863,21 @@ KWD::Window::resizeDecoration (bool force)
 	mPixmap = None;
     }
 
-    if (mPicture)
-    {
-	XRenderFreePicture (QX11Info::display(), mPicture);
-	mPicture = 0;
-    }
-
-    if (w != width() || h != height())
-    {
-	mPendingConfigure = 1;
-    }
-
-    setGeometry (QRect (mGeometry.x () + ROOT_OFF_X - mBorder.left,
-			mGeometry.y () + ROOT_OFF_Y - mBorder.top,
-			w, h));
-    XMoveResizeWindow (QX11Info::display(), winId(),
-		       mGeometry.x () + ROOT_OFF_X - mBorder.left,
-		       mGeometry.y () + ROOT_OFF_Y - mBorder.top,
-		       w, h);
-
-    mSize = QSize (w, h);
-
-    if (!mMapped)
-    {
-	mPendingMap = 1;
-
-	XReparentWindow (QX11Info::display(), winId (), mParentId, 0, 0);
-
-	show ();
-
-	mMapped = true;
-
-	if (mDamageId != winId ())
-	{
-	    mDamageId = winId ();
-	    XDamageCreate (QX11Info::display(), mDamageId,
-			   XDamageReportRawRectangles);
-	}
-    }
-
     mDecor->resize (QSize (w, h));
     mDecor->widget ()->show ();
     mDecor->widget ()->update ();
 
-    return TRUE;
-}
+    mPixmap = XCreatePixmap (QX11Info::display(),
+			     QX11Info::appRootWindow (),
+			     qMax (w, mGeometry.height ()),
+			     mExtents.top + mExtents.bottom + 
+			     mExtents.left + mExtents.right, 32);
 
-void
-KWD::Window::rebindPixmap (void)
-{
-    XRenderPictFormat *xformat;
+    mPixmapQt = QPixmap::fromX11Pixmap (mPixmap, QPixmap::ExplicitlyShared);
 
-    if (mPicture)
-	XRenderFreePicture (QX11Info::display(), mPicture);
+    mPixmapQt.fill (Qt::transparent);
 
-    if (mPixmap)
-	XFreePixmap (QX11Info::display(), mPixmap);
-
-    mPixmap = XCompositeNameWindowPixmap (QX11Info::display(), winId ());
-
-    xformat = XRenderFindVisualFormat (QX11Info::display(),
-				       (Visual *) QX11Info::appVisual ());
-
-    mPicture = XRenderCreatePicture (QX11Info::display(), mPixmap,
-				     xformat, 0, NULL);
-
-    updateShadow ();
-}
-
-bool
-KWD::Window::handleMap (void)
-{
-    if (!mPendingMap)
-	return FALSE;
-
-    mPendingMap = 0;
-
-    if (mPendingConfigure)
-	return FALSE;
-
-    rebindPixmap ();
-
-    return TRUE;
-}
-
-bool
-KWD::Window::handleConfigure (QSize size)
-{
-    if (!mPendingConfigure)
-	return FALSE;
-
-    if (size != mSize)
-	return FALSE;
-
-    mPendingConfigure = 0;
-    if (mPendingConfigure || mPendingMap)
-	return FALSE;
-
-    rebindPixmap ();
-
-    return TRUE;
+    mUpdateProperty = true;
 }
 
 void
@@ -1385,54 +895,54 @@ KWD::Window::updateBlurProperty (int topOffset,
     int     size = 0;
     int     w, h;
 
-    w = mGeometry.width () + mContext.extents.left + mContext.extents.right;
-    h = mGeometry.height () + mContext.extents.top + mContext.extents.bottom;
+    w = mGeometry.width () + mBorder.left + mBorder.right;
+    h = mGeometry.height () + mBorder.top + mBorder.bottom;
 
     if (blurType != BLUR_TYPE_NONE)
     {
 	QRegion r, shape = QRegion (0, 0, w, h);
 
 	if (mShapeSet)
-	    shape = mShape;
+	    shape = mShape.translated (-mPadding.left, -mPadding.top);
 
-	r = QRegion (0, 0, w, mContext.extents.top);
+	r = QRegion (0, 0, w, mBorder.top);
 	topQRegion = r.intersect (shape);
 	if (!topQRegion.isEmpty ())
 	{
-	    topQRegion.translate (-mContext.extents.left,
-				  -mContext.extents.top);
+	    topQRegion.translate (-mBorder.left,
+				  -mBorder.top);
 	    topRegion = topQRegion.handle ();
 	}
 
 	if (blurType == BLUR_TYPE_ALL)
 	{
-	    r = QRegion (0, h - mContext.extents.bottom,
-			 w, mContext.extents.bottom);
+	    r = QRegion (0, h - mBorder.bottom,
+			 w, mBorder.bottom);
 	    bottomQRegion = r.intersect (shape);
 	    if (!bottomQRegion.isEmpty ())
 	    {
-		bottomQRegion.translate (-mContext.extents.left,
-					 -(h - mContext.extents.bottom));
+		bottomQRegion.translate (-mBorder.left,
+					 -(h - mBorder.bottom));
 		bottomRegion = bottomQRegion.handle ();
 	    }
 
-	    r = QRegion (0, mContext.extents.top,
-			 mContext.extents.left, mGeometry.height ());
+	    r = QRegion (0, mBorder.top,
+			 mBorder.left, mGeometry.height ());
 	    leftQRegion = r.intersect (shape);
 	    if (!leftQRegion.isEmpty ())
 	    {
-		leftQRegion.translate (-mContext.extents.left,
-				       -mContext.extents.top);
+		leftQRegion.translate (-mBorder.left,
+				       -mBorder.top);
 		leftRegion = leftQRegion.handle ();
 	    }
 
-	    r = QRegion (w - mContext.extents.right, mContext.extents.top,
-			 mContext.extents.right, mGeometry.height ());
+	    r = QRegion (w - mBorder.right, mBorder.top,
+			 mBorder.right, mGeometry.height ());
 	    rightQRegion = r.intersect (shape);
 	    if (!rightQRegion.isEmpty ())
 	    {
-		rightQRegion.translate (-(w - mContext.extents.right),
-					-mContext.extents.top);
+		rightQRegion.translate (-(w - mBorder.right),
+					-mBorder.top);
 		rightRegion = rightQRegion.handle ();
 	    }
 	}
@@ -1481,10 +991,8 @@ KWD::Window::updateProperty (void)
     decor_extents_t maxExtents;
     long	    data[256];
     decor_quad_t    quads[N_QUADS_MAX];
-    int		    nQuad;
-    int		    lh, rh;
-    int		    w;
-    int		    minWidth;
+    int		    nQuad = 0;
+    int             left, right, top, bottom, width, height;
     unsigned int    saveState;
 
     if (mType == Default)
@@ -1499,29 +1007,22 @@ KWD::Window::updateProperty (void)
     mState = saveState;
     mDecor->borders (mBorder.left, mBorder.right, mBorder.top, mBorder.bottom);
 
-    if (mLayout.rotation)
-	lh = mLayout.left.x2 - mLayout.left.x1;
-    else
-	lh = mLayout.left.y2 - mLayout.left.y1;
-
-    if (mLayout.rotation)
-	rh = mLayout.right.x2 - mLayout.right.x1;
-    else
-	rh = mLayout.right.y2 - mLayout.right.y1;
-
-    w = mLayout.top.x2 - mLayout.top.x1 - mContext.left_space -
-	mContext.right_space;
+    left = mExtents.left;
+    right = mExtents.right;
+    top = mExtents.top;
+    bottom = mExtents.bottom;
+    width = mGeometry.width ();
+    height = mGeometry.height ();
 
     if (mType == Normal)
     {
-	int	topXOffset = w / 2;
+        decor_quad_t *q = quads;
+	int n = 0;
+
+	int	topXOffset = width;
 	QWidget *widget = mDecor->widget ();
 	int	x;
 
-	x = w - mContext.left_space - mContext.left_corner_space;
-	if (x > topXOffset)
-	    topXOffset = x;
-
 	if (widget)
 	{
 	    const QList<QObject*> children = widget->children ();
@@ -1535,34 +1036,78 @@ KWD::Window::updateProperty (void)
 
 		child = static_cast <QWidget *> (obj);
 
-		x = child->x () - mBorder.left - 2;
-		if (x > w / 2 && x < topXOffset)
+		x = child->x () - mExtents.left - 2;
+		if (x > width / 2 && x < topXOffset)
 		    topXOffset = x;
 	    }
 	}
 
-	nQuad = decor_set_lXrXtXbX_window_quads (quads,
-						 &mContext,
-						 &mLayout,
-						 lh / 2,
-						 rh / 2,
-						 topXOffset,
-						 w / 2);
+	// top quads
+	n = decor_set_horz_quad_line (q, left, topXOffset, right, 
+				      width - topXOffset - 1, -top, 0, GRAVITY_NORTH,
+				      left + right + width, -(width - topXOffset - 1),
+				      GRAVITY_EAST, 0, 0);
 
-	updateBlurProperty (topXOffset, w / 2, lh / 2, rh / 2);
+	q += n; nQuad += n;
+	
+	// bottom quads
+	n = decor_set_horz_quad_line (q, left, width / 2, right, (width / 2) - 1, 0,
+				      bottom, GRAVITY_SOUTH, left + right + width,
+				      -((width / 2) - 1), GRAVITY_EAST, 0, top);
 
-	minWidth = mContext.left_corner_space + 1 + mContext.right_corner_space;
+	q += n; nQuad += n;
+
+	// left quads
+	n = decor_set_vert_quad_row (q, 0, height / 2, 0, (height / 2) - 1, -left, 0,
+				     GRAVITY_WEST, height, -((height / 2) - 1),
+				     GRAVITY_SOUTH, 0, top + bottom, 1);
+
+	q += n; nQuad += n;
+
+	// right quads
+	n = decor_set_vert_quad_row (q, 0, height / 2, 0, (height / 2) - 1, 0, right,
+				     GRAVITY_EAST, height, -((height / 2) - 1),
+				     GRAVITY_SOUTH, 0, top + bottom + left, 1);
+
+	q += n; nQuad += n;
+
+	updateBlurProperty (topXOffset, width / 2, height / 2, height / 2);
     }
     else
     {
-	nQuad = decor_set_lSrStSbS_window_quads (quads, &mContext, &mLayout);
+	decor_quad_t *q = quads;
+	int n = 0;
+	
+	// top
+	n = decor_set_horz_quad_line (q, left, 0, right, 0, -top, 0,
+				      GRAVITY_NORTH, left + right + width,
+				      width / 2, 0, 0, 0);
 
-	minWidth = 1;
+        q += n; nQuad += n;
+	
+	// bottom
+	n = decor_set_horz_quad_line (q, left, 0, right, 0, 0, bottom,
+				      GRAVITY_SOUTH, left + right + width,
+				      width / 2, 0, 0, top);
+
+        q += n; nQuad += n;
+	
+	// left
+	n = decor_set_vert_quad_row (q, 0, 0, 0, 0, -left, 0, GRAVITY_WEST,
+				     height, height / 2, 0, 0, top + bottom, 1);
+
+	q += n; nQuad += n;
+
+	// right
+	n = decor_set_vert_quad_row (q, 0, 0, 0, 0, 0, right, GRAVITY_EAST,
+				     height, height / 2, 0, 0, top + bottom + left, 1);
+	
+	q += n; nQuad += n;
     }
 
-    decor_quads_to_property (data, mTexturePixmap.handle (),
+    decor_quads_to_property (data, mPixmap,
 			     &mBorder, &maxExtents,
-			     minWidth, 0,
+			     1, 0,
 			     quads, nQuad);
 
     KWD::trapXError ();
@@ -1606,7 +1151,13 @@ KWD::Window::updateSelected (WId selectedId)
 void
 KWD::Window::updateWindowGeometry (void)
 {
+    KWD::trapXError ();
     KWindowInfo wInfo = KWindowSystem::windowInfo (mClientId, NET::WMGeometry);
+    KWD::popXError ();
+
+    if (!wInfo.valid ())
+	return;
+
     QRect	geometry = wInfo.geometry ();
     int         w, h;
 
@@ -1617,18 +1168,13 @@ KWD::Window::updateWindowGeometry (void)
 	mGeometry.height () != geometry.height ())
     {
 	mGeometry = geometry;
-
-	if (resizeDecoration ())
-	    return;
+	resizeDecoration ();
     }
     else if (mGeometry.x ()  != geometry.x () ||
 	mGeometry.y () != geometry.y ())
     {
 	mGeometry = geometry;
     }
-
-    move (mGeometry.x () + ROOT_OFF_X - mBorder.left,
-	  mGeometry.y () + ROOT_OFF_Y - mBorder.top);
 }
 
 void
@@ -1637,14 +1183,10 @@ KWD::Window::reloadDecoration (void)
     delete mDecor;
     mDecor = 0;
 
-    mMapped   = false;
-    mShapeSet = false;
+    delete mPaintRedirector;
+    mPaintRedirector = 0;
 
-    if (mShadow)
-    {
-	decor_shadow_destroy (QX11Info::display(), mShadow);
-	mShadow = NULL;
-    }
+    mShapeSet = false;
 
     createDecoration ();
 }
@@ -1652,7 +1194,7 @@ KWD::Window::reloadDecoration (void)
 Cursor
 KWD::Window::positionToCursor (QPoint pos)
 {
-    switch (mDecor->mousePosition (pos)) {
+    switch (mDecor->mousePosition (pos + QPoint (mPadding.left, mPadding.top))) {
     case PositionCenter:
 	return cursors[1][1].cursor;
     case PositionLeft:
@@ -1924,6 +1466,9 @@ KWD::Window::moveWindow (QMouseEvent *qme)
     NET::Direction direction;
 
     direction = positionToDirection (mDecor->mousePosition (qme->pos ()));
+    
+    QPoint p (mGeometry.x () - mExtents.left, mGeometry.y () - mExtents.top);
+    p += qme->pos ();
 
     XUngrabPointer (QX11Info::display(), CurrentTime);
     XUngrabKeyboard (QX11Info::display(), CurrentTime);
@@ -1931,9 +1476,10 @@ KWD::Window::moveWindow (QMouseEvent *qme)
     Decorator::rootInfo ()->restackRequest (mClientId, NET::FromApplication,
 			 		    None, Above,
 					    QX11Info::appTime());
+					    
     Decorator::rootInfo ()->moveResizeRequest (mClientId,
-					       qme->globalX (),
-					       qme->globalY (),
+					       p.x (),
+					       p.y (),
 					       direction);
     mFakeRelease = true;
 
@@ -1962,14 +1508,8 @@ KWD::Window::performMouseCommand (Options::MouseCommand command,
 	setShade (false);
 	break;
     case Options::MouseOperationsMenu:
-    {
-	QPoint mp (0, 0);
-
-	if (qme)
-	    mp = mapToGlobal (qme->pos ());
-
-	showWindowMenu (mp);
-    } break;
+	showWindowMenu (mDecor->widget ()->mapToGlobal (qme->pos ()));
+	break;
     case Options::MouseMaximize:
 	maximize (KDecoration::MaximizeFull);
 	break;
@@ -2041,145 +1581,6 @@ KWD::Window::performMouseCommand (Options::MouseCommand command,
     }
 }
 
-void
-KWD::Window::processDamage (void)
-{
-    QRegion r1, r2;
-    int     xOff, yOff, w;
-    double  alpha;
-    int     shade_alpha;
-
-    if (isActive ())
-    {
-	alpha	    = activeDecorationOpacity;
-	shade_alpha = activeDecorationOpacityShade;
-    }
-    else
-    {
-	alpha	    = decorationOpacity;
-	shade_alpha = decorationOpacityShade;
-    }
-
-    if (!mPixmap)
-	return;
-
-    if (mDamage.isEmpty ())
-	return;
-
-    if (mShapeSet)
-	mDamage = mShape.intersect (mDamage);
-
-    w = mGeometry.width () + mContext.extents.left + mContext.extents.right;
-
-    xOff = 0;
-    yOff = 0;
-
-    r1 = QRegion (xOff, yOff, w, mContext.extents.top);
-    r2 = r1.intersect (mDamage);
-
-    if (!r2.isEmpty ())
-    {
-	r2.translate (-xOff, -yOff);
-
-	decor_blend_border_picture (QX11Info::display(),
-				    &mContext,
-				    mPicture,
-				    xOff, xOff,
-				    mTexturePicture,
-				    &mLayout,
-				    BORDER_TOP,
-				    r2.handle (),
-				    (unsigned short) (alpha * 0xffff),
-				    shade_alpha,
-				    TRUE);
-    }
-
-    xOff = 0;
-    yOff = mContext.extents.top + mGeometry.height ();
-
-    r1 = QRegion (xOff, yOff, w, mContext.extents.bottom);
-    r2 = r1.intersect (mDamage);
-
-    if (!r2.isEmpty ())
-    {
-	r2.translate (-xOff, -yOff);
-
-	decor_blend_border_picture (QX11Info::display(),
-				    &mContext,
-				    mPicture,
-				    xOff, yOff,
-				    mTexturePicture,
-				    &mLayout,
-				    BORDER_BOTTOM,
-				    r2.handle (),
-				    (unsigned short) (alpha * 0xffff),
-				    shade_alpha,
-				    TRUE);
-    }
-
-    xOff = 0;
-    yOff = mContext.extents.top;
-
-    r1 = QRegion (xOff, yOff, mContext.extents.left, mGeometry.height ());
-    r2 = r1.intersect (mDamage);
-
-    if (!r2.isEmpty ())
-    {
-	r2.translate (-xOff, -yOff);
-
-	decor_blend_border_picture (QX11Info::display(),
-				    &mContext,
-				    mPicture,
-				    xOff, yOff,
-				    mTexturePicture,
-				    &mLayout,
-				    BORDER_LEFT,
-				    r2.handle (),
-				    (unsigned short) (alpha * 0xffff),
-				    shade_alpha,
-				    TRUE);
-    }
-
-    xOff = mContext.extents.left + mGeometry.width ();
-    yOff = mContext.extents.top;
-
-    r1 = QRegion (xOff, yOff, mContext.extents.right, mGeometry.height ());
-    r2 = r1.intersect (mDamage);
-
-    if (!r2.isEmpty ())
-    {
-	r2.translate (-xOff, -yOff);
-
-	decor_blend_border_picture (QX11Info::display(),
-				    &mContext,
-				    mPicture,
-				    xOff, yOff,
-				    mTexturePicture,
-				    &mLayout,
-				    BORDER_RIGHT,
-				    r2.handle (),
-				    (unsigned short) (alpha * 0xffff),
-				    shade_alpha,
-				    TRUE);
-    }
-
-    mDamage = QRegion ();
-
-    XRenderComposite (QX11Info::display(),
-		      PictOpSrc,
-		      mTexturePicture,
-		      None,
-		      mDecorationPicture,
-		      0, 0,
-		      0, 0,
-		      0, 0,
-		      mTexturePixmap.width (),
-		      mTexturePixmap.height ());
-
-    if (mUpdateProperty)
-	updateProperty ();
-}
-
 void
 KWD::Window::showKillProcessDialog (Time timestamp)
 {
@@ -2223,3 +1624,163 @@ KWD::Window::hideKillProcessDialog (void)
 	mProcessKiller.terminate ();
     }
 }
+
+void
+KWD::Window::decorRepaintPending ()
+{
+    if (!mPaintRedirector || !mPixmap)
+        return;
+
+    QRegion reg = mPaintRedirector->pendingRegion();
+    if (reg.isEmpty())
+        return;
+    
+    QRect bBox = reg.boundingRect();
+ 
+    if (mShapeSet)
+      reg &= mShape;
+    
+    int l = mExtents.left;
+    int r = mExtents.right;
+    int t = mExtents.top;
+    int b = mExtents.bottom;
+    int w = mGeometry.width ();
+    int h = mGeometry.height ();
+    
+    QRect top = QRect (0, 0, w + l + r, t);
+    QRect bottom = QRect (0, t + h, w + l + r, b);
+    QRect left = QRect (0, t, l, h);
+    QRect right = QRect (l + w, t, r, h);
+        
+    QRegion rtop = reg & top;
+    QRegion rbottom = reg & bottom;
+    QRegion rleft = reg & left;
+    QRegion rright = reg & right;
+
+    QPixmap p = mPaintRedirector->performPendingPaint();
+
+    QPainter pt (&mPixmapQt);
+    pt.setCompositionMode( QPainter::CompositionMode_Source );
+    
+    QRect bb, pb;
+ 
+    // Top
+    if (!rtop.isEmpty ())
+    {
+      bb = rtop.boundingRect();
+      pb = bb;
+      pb.moveTo (bb.topLeft () - bBox.topLeft ());
+      pt.resetTransform ();
+      pt.setClipRegion( reg );
+      pt.drawPixmap( bb.topLeft(), p, pb );
+    }
+    
+    // Bottom
+    if (!rbottom.isEmpty ())
+    {
+      bb = rbottom.boundingRect();
+      pb = bb;
+      pb.moveTo (bb.topLeft () - bBox.topLeft ());
+      pt.resetTransform ();
+      pt.translate(0, -h);
+      pt.setClipRegion( reg );
+      pt.drawPixmap( bb.topLeft(), p, pb );
+    }
+    
+    // Left
+    if (!rleft.isEmpty ())
+    {
+      bb = rleft.boundingRect();
+      pb = bb;
+      pb.moveTo (bb.topLeft () - bBox.topLeft ());
+      pt.resetTransform ();
+      pt.translate(0, t + b);
+      pt.rotate (90);
+      pt.scale (1.0, -1.0);
+      pt.translate(0, -t);
+      pt.setClipRegion( reg );
+      pt.drawPixmap( bb.topLeft(), p, pb );
+    }
+    
+    // Right
+    if (!rright.isEmpty ())
+    {
+      bb = rright.boundingRect();
+      pb = bb;
+      pb.moveTo (bb.topLeft () - bBox.topLeft ());
+      pt.resetTransform ();
+      pt.translate(0, t + b + l);
+      pt.rotate (90);
+      pt.scale (1.0, -1.0);
+      pt.translate(- (l + w), -t);
+      pt.setClipRegion( reg );
+      pt.drawPixmap( bb.topLeft(), p, pb );
+    }
+
+    
+
+    if (mUpdateProperty)
+	updateProperty ();
+}
+
+QWidget *
+KWD::Window::decorWidget (void) const
+{
+    if (!mDecor)
+	return 0;
+    return mDecor->widget ();
+}
+
+QWidget *
+KWD::Window::childAt (int x, int y) const
+{
+    if (!mDecor)
+	return 0;
+
+    QWidget *child = mDecor->widget ()->childAt (x + mPadding.left, y + mPadding.top);
+    return (child)? child : decorWidget ();
+}
+
+QPoint 
+KWD::Window::mapToChildAt (QPoint p) const
+{
+    if (!mDecor)
+	return p;
+    if (childAt (p.x (), p.y ()) == decorWidget ())
+	return p + QPoint (mPadding.left, mPadding.right);
+    return childAt (p.x (), p.y ())->mapFrom (decorWidget (), p + QPoint (mPadding.left, mPadding.right));
+}
+
+bool 
+KWD::Window::eventFilter (QObject* o, QEvent* e)
+{
+    if (mDecor == NULL || o != mDecor->widget ())
+	return false;
+    if (e->type() == QEvent::Resize)
+    {
+	QResizeEvent* ev = static_cast<QResizeEvent*> (e);
+	// Filter out resize events that inform about size different than frame size.
+	// This will ensure that mDecor->width() etc. and mDecor->widget()->width() will be in sync.
+	// These events only seem to be delayed events from initial resizing before show() was called
+	// on the decoration widget.
+	if (ev->size () != (mGeometry.size () + QSize (mExtents.left + mExtents.right, 
+						       mExtents.top + mExtents.bottom)))
+	{
+	    int w = mGeometry.width () + mExtents.left + mExtents.right;
+	    int h = mGeometry.height () + mExtents.top + mExtents.bottom;
+    
+	    mDecor->resize (QSize (w, h));
+	    return true;
+	}
+	// HACK: Avoid decoration redraw delays. On resize Qt sets WA_WStateConfigPending
+	// which delays all painting until a matching ConfigureNotify event comes.
+	// But this process itself is the window manager, so it's not needed
+	// to wait for that event, the geometry is known.
+	// Note that if Qt in the future changes how this flag is handled and what it
+	// triggers then this may potentionally break things. See mainly QETWidget::translateConfigEvent().
+	mDecor->widget()->setAttribute( Qt::WA_WState_ConfigPending, false );
+	mDecor->widget()->update();
+	return false;
+    }
+    return false;
+}
