@@ -102,8 +102,8 @@ typedef struct _SwitchScreen {
 
     Window popupWindow;
 
-    Window	 selectedWindow;
-    Window	 zoomedWindow;
+    CompWindow	 *selectedWindow;
+    CompWindow	 *zoomedWindow;
     unsigned int lastActiveNum;
 
     float zoom;
@@ -151,7 +151,8 @@ typedef struct {
 
 #define BOX_WIDTH 3
 
-#define ICON_SIZE 64
+#define ICON_SIZE 48
+#define MAX_ICON_SIZE 256
 
 static float _boxVertices[] =
 {
@@ -247,12 +248,17 @@ switchSetScreenOption (CompPlugin      *plugin,
 static void
 setSelectedWindowHint (CompScreen *s)
 {
+    Window selectedWindowId = None;
+
     SWITCH_DISPLAY (s->display);
     SWITCH_SCREEN (s);
 
+    if (ss->selectedWindow && !ss->selectedWindow->destroyed)
+	selectedWindowId = ss->selectedWindow->id;
+
     XChangeProperty (s->display->display, ss->popupWindow, sd->selectWinAtom,
 		     XA_WINDOW, 32, PropModeReplace,
-		     (unsigned char *) &ss->selectedWindow, 1);
+		     (unsigned char *) &selectedWindowId, 1);
 }
 
 static Bool
@@ -260,6 +266,9 @@ isSwitchWin (CompWindow *w)
 {
     SWITCH_SCREEN (w->screen);
 
+    if (w->destroyed)
+	return FALSE;
+
     if (!w->mapNum || w->attrib.map_state != IsViewable)
     {
 	if (ss->opt[SWITCH_SCREEN_OPTION_MINIMIZED].value.b)
@@ -390,7 +399,7 @@ switchUpdateWindowList (CompScreen *s,
     ss->pos  = ((count >> 1) - ss->nWindows) * WIDTH;
     ss->move = 0;
 
-    ss->selectedWindow = ss->windows[0]->id;
+    ss->selectedWindow = ss->windows[0];
 
     x = s->outputDev[s->currentOutputDev].region.extents.x1 +
 	s->outputDev[s->currentOutputDev].width / 2;
@@ -446,7 +455,7 @@ switchToWindow (CompScreen *s,
 
     for (cur = 0; cur < ss->nWindows; cur++)
     {
-	if (ss->windows[cur]->id == ss->selectedWindow)
+	if (ss->windows[cur] == ss->selectedWindow)
 	    break;
     }
 
@@ -460,7 +469,7 @@ switchToWindow (CompScreen *s,
 
     if (w)
     {
-	Window old = ss->selectedWindow;
+	CompWindow *old = ss->selectedWindow;
 
 	if (ss->selection == AllViewports &&
 	    ss->opt[SWITCH_SCREEN_OPTION_AUTO_ROTATE].value.b)
@@ -489,12 +498,12 @@ switchToWindow (CompScreen *s,
 	}
 
 	ss->lastActiveNum  = w->activeNum;
-	ss->selectedWindow = w->id;
+	ss->selectedWindow = w;
 
 	if (!ss->zoomedWindow)
 	    ss->zoomedWindow = ss->selectedWindow;
 
-	if (old != w->id)
+	if (old != w)
 	{
 	    if (toNext)
 		ss->move -= WIDTH;
@@ -517,12 +526,8 @@ switchToWindow (CompScreen *s,
 
 	addWindowDamage (w);
 
-	if (old)
-	{
-	    w = findWindowAtScreen (s, old);
-	    if (w)
-		addWindowDamage (w);
-	}
+	if (old && !old->destroyed)
+	    addWindowDamage (old);
     }
 }
 
@@ -590,11 +595,11 @@ switchInitiate (CompScreen            *s,
 
     SWITCH_SCREEN (s);
 
-    if (otherScreenGrabExist (s, "switcher", "scale", "cube", 0))
+    if (otherScreenGrabExist (s, "cube", NULL))
 	return;
 
     ss->selection      = selection;
-    ss->selectedWindow = None;
+    ss->selectedWindow = NULL;
 
     count = switchCountWindows (s);
     if (count < 1)
@@ -752,24 +757,20 @@ switchTerminate (CompDisplay     *d,
 
 	    if (state & CompActionStateCancel)
 	    {
-		ss->selectedWindow = None;
-		ss->zoomedWindow   = None;
+		ss->selectedWindow = NULL;
+		ss->zoomedWindow   = NULL;
 	    }
 
-	    if (state && ss->selectedWindow)
-	    {
-		w = findWindowAtScreen (s, ss->selectedWindow);
-		if (w)
-		    sendWindowActivationRequest (w->screen, w->id);
-	    }
+	    if (state && ss->selectedWindow && !ss->selectedWindow->destroyed)
+		sendWindowActivationRequest (s, ss->selectedWindow->id);
 
 	    removeScreenGrab (s, ss->grabIndex, 0);
 	    ss->grabIndex = 0;
 
 	    if (!ss->zooming)
 	    {
-		ss->selectedWindow = None;
-		ss->zoomedWindow   = None;
+		ss->selectedWindow = NULL;
+		ss->zoomedWindow   = NULL;
 
 		switchActivateEvent (s, FALSE);
 	    }
@@ -778,7 +779,7 @@ switchTerminate (CompDisplay     *d,
 		ss->moreAdjust = 1;
 	    }
 
-	    ss->selectedWindow = None;
+	    ss->selectedWindow = NULL;
 	    setSelectedWindowHint (s);
 
 	    ss->lastActiveNum = 0;
@@ -922,16 +923,14 @@ switchPrevPanel (CompDisplay     *d,
 
 static void
 switchWindowRemove (CompDisplay *d,
-		    Window	id)
+		    CompWindow  *w)
 {
-    CompWindow *w;
-
-    w = findWindowAtDisplay (d, id);
     if (w)
     {
 	Bool   inList = FALSE;
 	int    count, j, i = 0;
-	Window selected, old;
+	CompWindow *selected;
+	CompWindow *old;
 
 	SWITCH_SCREEN (w->screen);
 
@@ -946,12 +945,12 @@ switchWindowRemove (CompDisplay *d,
 	    {
 		inList = TRUE;
 
-		if (w->id == selected)
+		if (w == selected)
 		{
-		    if (i < ss->nWindows)
-			selected = ss->windows[i + 1]->id;
+		    if (i + 1 < ss->nWindows)
+			selected = ss->windows[i + 1];
 		    else
-			selected = ss->windows[0]->id;
+			selected = ss->windows[0];
 		}
 
 		ss->nWindows--;
@@ -1002,7 +1001,7 @@ switchWindowRemove (CompDisplay *d,
 
 	for (i = 0; i < ss->nWindows; i++)
 	{
-	    ss->selectedWindow = ss->windows[i]->id;
+	    ss->selectedWindow = ss->windows[i];
 
 	    if (ss->selectedWindow == selected)
 		break;
@@ -1025,13 +1024,13 @@ switchWindowRemove (CompDisplay *d,
 
 	if (old != ss->selectedWindow)
 	{
+	    ss->zoomedWindow = ss->selectedWindow;
+
+	    addWindowDamage (ss->selectedWindow);
 	    addWindowDamage (w);
 
-	    w = findWindowAtScreen (w->screen, old);
-	    if (w)
-		addWindowDamage (w);
-
-	    ss->moreAdjust = 1;
+	    if (old && !old->destroyed)
+		addWindowDamage (old);
 	}
     }
 }
@@ -1085,10 +1084,11 @@ static void
 switchHandleEvent (CompDisplay *d,
 		   XEvent      *event)
 {
-    CompWindow *w;
+    CompWindow *w = NULL;
     SWITCH_DISPLAY (d);
 
     switch (event->type) {
+    case MapNotify:
 	w = findWindowAtDisplay (d, event->xmap.window);
 	if (w)
 	{
@@ -1103,6 +1103,13 @@ switchHandleEvent (CompDisplay *d,
 	    }
 	}
 	break;
+    case DestroyNotify:
+	/* We need to get the CompWindow * for event->xdestroywindow.window
+	   here because in the (*d->handleEvent) call below, that CompWindow's
+	   id will become 1, so findWindowAtDisplay won't be able to find the
+	   CompWindow after that. */
+	w = findWindowAtDisplay (d, event->xdestroywindow.window);
+	break;
     }
 
     UNWRAP (sd, d, handleEvent);
@@ -1111,10 +1118,11 @@ switchHandleEvent (CompDisplay *d,
 
     switch (event->type) {
     case UnmapNotify:
-	switchWindowRemove (d, event->xunmap.window);
+	w = findWindowAtDisplay (d, event->xunmap.window);
+	switchWindowRemove (d, w);
 	break;
     case DestroyNotify:
-	switchWindowRemove (d, event->xdestroywindow.window);
+	switchWindowRemove (d, w);
 	break;
     case PropertyNotify:
 	if (event->xproperty.atom == sd->selectFgColorAtom)
@@ -1245,8 +1253,8 @@ switchPreparePaintScreen (CompScreen *s,
 			ss->translate  = 0.0f;
 			ss->sTranslate = ss->zoom;
 
-			ss->selectedWindow = None;
-			ss->zoomedWindow   = None;
+			ss->selectedWindow = NULL;
+			ss->zoomedWindow   = NULL;
 
 			if (ss->grabIndex)
 			{
@@ -1331,12 +1339,13 @@ switchPaintOutput (CompScreen		   *s,
 
 	if (ss->opt[SWITCH_SCREEN_OPTION_BRINGTOFRONT].value.b)
 	{
-	    zoomed = findWindowAtScreen (s, ss->zoomedWindow);
-	    if (zoomed)
+	    zoomed = ss->zoomedWindow;
+	    if (zoomed && !zoomed->destroyed)
 	    {
 		CompWindow *w;
 
-		for (w = zoomed->prev; w && w->id <= 1; w = w->prev);
+		for (w = zoomed->prev; w && w->id <= 1; w = w->prev)
+		    ;
 		zoomedAbove = (w) ? w->id : None;
 
 		unhookWindowFromScreen (s, zoomed);
@@ -1465,6 +1474,7 @@ switchPaintThumb (CompWindow		  *w,
 	FragmentAttrib	      fragment;
 	CompTransform	      wTransform = *transform;
 	int		      ww, wh;
+	GLenum                filter;
 
 	SWITCH_SCREEN (w->screen);
 
@@ -1513,6 +1523,11 @@ switchPaintThumb (CompWindow		  *w,
 	glPushMatrix ();
 	glLoadMatrixf (wTransform.m);
 
+	filter = w->screen->display->textureFilter;
+
+	if (ss->opt[SWITCH_SCREEN_OPTION_MIPMAP].value.b)
+	    w->screen->display->textureFilter = GL_LINEAR_MIPMAP_LINEAR;
+
 	/* XXX: replacing the addWindowGeometry function like this is
 	   very ugly but necessary until the vertex stage has been made
 	   fully pluggable. */
@@ -1522,17 +1537,30 @@ switchPaintThumb (CompWindow		  *w,
 				 mask);
 	w->screen->addWindowGeometry = oldAddWindowGeometry;
 
+	w->screen->display->textureFilter = filter;
+
 	glPopMatrix ();
 
 	if (ss->opt[SWITCH_SCREEN_OPTION_ICON].value.b)
 	{
-	    icon = getWindowIcon (w, ICON_SIZE, ICON_SIZE);
+	    icon = getWindowIcon (w, MAX_ICON_SIZE, MAX_ICON_SIZE);
 	    if (icon)
 	    {
-		sAttrib.xScale = sAttrib.yScale = 1.0f;
+		sAttrib.xScale = (float) ICON_SIZE / icon->width;
+		sAttrib.yScale = (float) ICON_SIZE / icon->height;
+		/*
+		sAttrib.xScale =
+		    ((WIDTH - (SPACE << 1)) / 2.5f) / icon->width;
+		sAttrib.yScale =
+		    ((HEIGHT - (SPACE << 1)) / 2.5f) / icon->height;
+		*/
+		if (sAttrib.xScale < sAttrib.yScale)
+		    sAttrib.yScale = sAttrib.xScale;
+		else
+		    sAttrib.xScale = sAttrib.yScale;
 
-		wx = x + WIDTH  - icon->width  - SPACE;
-		wy = y + HEIGHT - icon->height - SPACE;
+		wx = x + WIDTH  - icon->width  * sAttrib.xScale - SPACE;
+		wy = y + HEIGHT - icon->height * sAttrib.yScale - SPACE;
 	    }
 	}
     }
@@ -1541,26 +1569,19 @@ switchPaintThumb (CompWindow		  *w,
 	width  = WIDTH  - (WIDTH  >> 2);
 	height = HEIGHT - (HEIGHT >> 2);
 
-	icon = getWindowIcon (w, width, height);
+	icon = getWindowIcon (w, MAX_ICON_SIZE, MAX_ICON_SIZE);
 	if (!icon)
 	    icon = w->screen->defaultIcon;
 
 	if (icon)
 	{
-	    int iw, ih;
+	    float iw, ih;
 
 	    iw = width  - SPACE;
 	    ih = height - SPACE;
 
-	    if (icon->width < (iw >> 1))
-		sAttrib.xScale = (iw / icon->width);
-	    else
-		sAttrib.xScale = 1.0f;
-
-	    if (icon->height < (ih >> 1))
-		sAttrib.yScale = (ih / icon->height);
-	    else
-		sAttrib.yScale = 1.0f;
+	    sAttrib.xScale = (iw / icon->width);
+	    sAttrib.yScale = (ih / icon->height);
 
 	    if (sAttrib.xScale < sAttrib.yScale)
 		sAttrib.yScale = sAttrib.xScale;
@@ -1640,7 +1661,6 @@ switchPaintWindow (CompWindow		   *w,
 
     if (w->id == ss->popupWindow)
     {
-	GLenum         filter;
 	int            x, y, x1, x2, cx, i;
 	unsigned short color[4];
 
@@ -1660,11 +1680,6 @@ switchPaintWindow (CompWindow		   *w,
 	x = x1 + ss->pos;
 	y = w->attrib.y + SPACE;
 
-	filter = s->display->textureFilter;
-
-	if (ss->opt[SWITCH_SCREEN_OPTION_MIPMAP].value.b)
-	    s->display->textureFilter = GL_LINEAR_MIPMAP_LINEAR;
-
 	glPushAttrib (GL_SCISSOR_BIT);
 
 	glEnable (GL_SCISSOR_TEST);
@@ -1692,8 +1707,6 @@ switchPaintWindow (CompWindow		   *w,
 
 	glPopAttrib ();
 
-	s->display->textureFilter = filter;
-
 	cx = w->attrib.x + (w->width >> 1);
 
 	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
@@ -1711,7 +1724,7 @@ switchPaintWindow (CompWindow		   *w,
 	glDisable (GL_BLEND);
 	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
     }
-    else if (w->id == ss->selectedWindow)
+    else if (w == ss->selectedWindow)
     {
 	if (ss->opt[SWITCH_SCREEN_OPTION_BRINGTOFRONT].value.b &&
 	    ss->selectedWindow == ss->zoomedWindow)
@@ -1746,7 +1759,7 @@ switchPaintWindow (CompWindow		   *w,
 	}
 
 	if (ss->opt[SWITCH_SCREEN_OPTION_BRINGTOFRONT].value.b &&
-	    w->id == ss->zoomedWindow)
+	    w == ss->zoomedWindow)
 	    zoomType = ZOOMED_WINDOW_MASK;
 
 	if (!(ss->zoomMask & zoomType))
@@ -1951,8 +1964,8 @@ switchInitScreen (CompPlugin *p,
 
     ss->popupWindow = None;
 
-    ss->selectedWindow = None;
-    ss->zoomedWindow   = None;
+    ss->selectedWindow = NULL;
+    ss->zoomedWindow   = NULL;
 
     ss->lastActiveNum  = 0;
 
