@@ -673,26 +673,81 @@ setDisplayOption (CompPlugin	  *plugin,
 static void
 updatePlugins (CompDisplay *d)
 {
-    CompOption *o;
-    CompPlugin *p, **pop = 0;
-    int	       nPop, i, j;
+    CompOption      *o;
+    CompPlugin      *p, **pop = 0;
+    int	            nPop, i, j, k;
+    CompOptionValue *pList;
+    int             pListCount = 1;
 
     d->dirtyPluginList = FALSE;
 
     o = &d->opt[COMP_DISPLAY_OPTION_ACTIVE_PLUGINS];
 
-    /* The old plugin list always begins with the core plugin. To make sure
-       we don't unnecessarily unload plugins if the new plugin list does not
-       contain the core plugin, we have to use an offset */
-    if (o->value.list.nValue > 0 && strcmp (o->value.list.value[0].s, "core"))
-	i = 0;
-    else
-	i = 1;
+    /* determine number of plugins, which is core + initial plugins +
+       plugins in option list additional to initial plugins */
+    for (i = 0; i < nInitialPlugins; i++)
+	if (strcmp (initialPlugins[i], "core") != 0)
+	    pListCount++;
+
+    for (i = 0; i < o->value.list.nValue; i++)
+    {
+	if (strcmp (o->value.list.value[i].s, "core") == 0)
+	    continue;
+
+	for (j = 0; j < nInitialPlugins; j++)
+	{
+	    if (strcmp (o->value.list.value[i].s, initialPlugins[j]) == 0)
+		break;
+	}
+
+	/* plugin is not in initial plugin list */
+	if (j == nInitialPlugins)
+	    pListCount++;
+    }
+
+    pList = malloc (sizeof (CompOptionValue) * pListCount);
+    if (!pList)
+    {
+	(*core.setOptionForPlugin) (&d->base, "core", o->name, &d->plugin);
+	return;
+    }
+
+    /* new plugin list needs core as first plugin */
+    pList[0].s = "core";
+
+    /* afterwards, add the initial plugins */
+    for (j = 0, k = 1; j < nInitialPlugins; j++)
+    {
+	/* avoid adding core twice */
+	if (strcmp (initialPlugins[j], "core") == 0)
+	    continue;
+	
+	pList[k++].s = initialPlugins[j];
+    }
+    j = k;
+
+    /* then add the plugins not in the initial plugin list */
+    for (i = 0; i < o->value.list.nValue; i++)
+    {
+	if (strcmp (o->value.list.value[i].s, "core") == 0)
+	    continue;
+
+	for (k = 0; k < nInitialPlugins; k++)
+	{
+	    if (strcmp (o->value.list.value[i].s, initialPlugins[k]) == 0)
+		break;
+	}
+	
+	if (k == nInitialPlugins)
+	    pList[j++].s = o->value.list.value[i].s;
+    }
+
+    assert (j == pListCount);
 
     /* j is initialized to 1 to make sure we never pop the core plugin */
-    for (j = 1; j < d->plugin.list.nValue && i < o->value.list.nValue; i++, j++)
+    for (i = j = 1; j < d->plugin.list.nValue && i < pListCount; i++, j++)
     {
-	if (strcmp (d->plugin.list.value[j].s, o->value.list.value[i].s))
+	if (strcmp (d->plugin.list.value[j].s, pList[i].s))
 	    break;
     }
 
@@ -704,6 +759,7 @@ updatePlugins (CompDisplay *d)
 	if (!pop)
 	{
 	    (*core.setOptionForPlugin) (&d->base, "core", o->name, &d->plugin);
+	    free (pList);
 	    return;
 	}
     }
@@ -715,13 +771,12 @@ updatePlugins (CompDisplay *d)
 	free (d->plugin.list.value[d->plugin.list.nValue].s);
     }
 
-    for (; i < o->value.list.nValue; i++)
+    for (; i < pListCount; i++)
     {
 	p = 0;
 	for (j = 0; j < nPop; j++)
 	{
-	    if (pop[j] && strcmp (pop[j]->vTable->name,
-				  o->value.list.value[i].s) == 0)
+	    if (pop[j] && strcmp (pop[j]->vTable->name, pList[i].s) == 0)
 	    {
 		if (pushPlugin (pop[j]))
 		{
@@ -734,7 +789,7 @@ updatePlugins (CompDisplay *d)
 
 	if (p == 0)
 	{
-	    p = loadPlugin (o->value.list.value[i].s);
+	    p = loadPlugin (pList[i].s);
 	    if (p)
 	    {
 		if (!pushPlugin (p))
@@ -775,6 +830,7 @@ updatePlugins (CompDisplay *d)
     if (nPop)
 	free (pop);
 
+    free (pList);
     (*core.setOptionForPlugin) (&d->base, "core", o->name, &d->plugin);
 }
 
@@ -2388,6 +2444,18 @@ removeDisplay (CompDisplay *d)
 
     objectFiniPlugins (&d->base);
 
+    if (d->edgeDelayHandle)
+    {
+	void *closure;
+
+	closure = compRemoveTimeout (d->edgeDelayHandle);
+	if (closure)
+	    free (closure);
+    }
+
+    if (d->autoRaiseHandle)
+	compRemoveTimeout (d->autoRaiseHandle);
+
     compRemoveTimeout (d->pingHandle);
 
     if (d->snDisplay)
