@@ -29,6 +29,7 @@
 #include <libxml/tree.h>
 #include <libxml/xpath.h>
 #include <libxml/xpathInternals.h>
+#include <locale.h>
 
 #include <compiz-core.h>
 
@@ -165,7 +166,7 @@ compAddMetadataFromFile (CompMetadata *metadata,
     status |= addMetadataFromFilename (metadata, METADATADIR, file);
     if (!status)
     {
-	compLogMessage (NULL, "core", CompLogLevelWarn,
+	compLogMessage ("core", CompLogLevelWarn,
 			"Unable to parse XML metadata from file \"%s%s\"",
 			file, EXTENSION);
 
@@ -184,7 +185,7 @@ compAddMetadataFromString (CompMetadata *metadata,
     doc = xmlReadMemory (string, strlen (string), NULL, NULL, 0);
     if (!doc)
     {
-	compLogMessage (NULL, "core", CompLogLevelWarn,
+	compLogMessage ("core", CompLogLevelWarn,
 			"Unable to parse XML metadata");
 
 	return FALSE;
@@ -214,7 +215,7 @@ compAddMetadataFromIO (CompMetadata	     *metadata,
     doc = xmlReadIO (ioread, ioclose, ioctx, NULL, NULL, 0);
     if (!doc)
     {
-	compLogMessage (NULL, "core", CompLogLevelWarn,
+	compLogMessage ("core", CompLogLevelWarn,
 			"Unable to parse XML metadata");
 
 	return FALSE;
@@ -470,12 +471,15 @@ initFloatValue (CompOptionValue	      *v,
 		xmlNodePtr	      node)
 {
     xmlChar *value;
+    char *loc;
 
     v->f = (r->f.min + r->f.max) / 2;
 
     if (!doc)
 	return;
 
+    loc = setlocale (LC_NUMERIC, NULL);
+    setlocale (LC_NUMERIC, "C");
     value = xmlNodeListGetString (doc, node->xmlChildrenNode, 1);
     if (value)
     {
@@ -486,6 +490,7 @@ initFloatValue (CompOptionValue	      *v,
 
 	xmlFree (value);
     }
+    setlocale (LC_NUMERIC, loc);
 }
 
 static void
@@ -866,11 +871,14 @@ initFloatRestriction (CompMetadata	    *metadata,
 		      const char	    *path)
 {
     char *value;
+    char *loc;
 
     r->f.min	   = MINSHORT;
     r->f.max	   = MAXSHORT;
     r->f.precision = 0.1f;
 
+    loc = setlocale (LC_NUMERIC, NULL);
+    setlocale (LC_NUMERIC, "C");
     value = stringFromMetadataPathElement (metadata, path, "min");
     if (value)
     {
@@ -891,10 +899,13 @@ initFloatRestriction (CompMetadata	    *metadata,
 	r->f.precision = strtod ((char *) value, NULL);
 	free (value);
     }
+
+    setlocale (LC_NUMERIC, loc);
 }
 
 static void
 initActionState (CompMetadata    *metadata,
+		 CompOptionType  type,
 		 CompActionState *state,
 		 const char      *path)
 {
@@ -923,6 +934,20 @@ initActionState (CompMetadata    *metadata,
 	free (grab);
     }
 
+    if (type == CompOptionTypeEdge)
+    {
+	char *noEdgeDelay;
+
+	noEdgeDelay = stringFromMetadataPathElement (metadata, path, "nodelay");
+	if (noEdgeDelay)
+	{
+	    if (strcmp (noEdgeDelay, "true") == 0)
+		*state |= CompActionStateNoEdgeDelay;
+
+	    free (noEdgeDelay);
+	}
+    }
+
     if (!initXPathFromMetadataPathElement (&xPath, metadata, BAD_CAST path,
 					   BAD_CAST "allowed"))
 	return;
@@ -933,8 +958,12 @@ initActionState (CompMetadata    *metadata,
 
 	value = xmlGetProp (*xPath.obj->nodesetval->nodeTab,
 			    BAD_CAST map[i].name);
-	if (value && xmlStrcmp (value, BAD_CAST "true") == 0)
-	    *state |= map[i].state;
+	if (value)
+	{
+	    if (xmlStrcmp (value, BAD_CAST "true") == 0)
+		*state |= map[i].state;
+	    xmlFree (value);
+	}
     }
 
     finiXPath (&xPath);
@@ -1006,23 +1035,23 @@ initOptionFromMetadataPath (CompDisplay   *d,
 	initColorValue (&option->value, defaultDoc, defaultNode);
 	break;
     case CompOptionTypeAction:
-	initActionState (metadata, &state, (char *) path);
+	initActionState (metadata, option->type, &state, (char *) path);
 	initActionValue (d, &option->value, state, defaultDoc, defaultNode);
 	break;
     case CompOptionTypeKey:
-	initActionState (metadata, &state, (char *) path);
+	initActionState (metadata, option->type, &state, (char *) path);
 	initKeyValue (d, &option->value, state, defaultDoc, defaultNode);
 	break;
     case CompOptionTypeButton:
-	initActionState (metadata, &state, (char *) path);
+	initActionState (metadata, option->type, &state, (char *) path);
 	initButtonValue (d, &option->value, state, defaultDoc, defaultNode);
 	break;
     case CompOptionTypeEdge:
-	initActionState (metadata, &state, (char *) path);
+	initActionState (metadata, option->type, &state, (char *) path);
 	initEdgeValue (d, &option->value, state, defaultDoc, defaultNode);
 	break;
     case CompOptionTypeBell:
-	initActionState (metadata, &state, (char *) path);
+	initActionState (metadata, option->type, &state, (char *) path);
 	initBellValue (d, &option->value, state, defaultDoc, defaultNode);
 	break;
     case CompOptionTypeMatch:
@@ -1054,7 +1083,8 @@ initOptionFromMetadataPath (CompDisplay   *d,
 	case CompOptionTypeButton:
 	case CompOptionTypeEdge:
 	case CompOptionTypeBell:
-	    initActionState (metadata, &state, (char *) path);
+	    initActionState (metadata, option->value.list.type,
+			     &state, (char *) path);
 	    break;
 	case CompOptionTypeMatch:
 	    helper = boolFromMetadataPathElement (metadata, (char *) path,
